# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.api.definitions import provider_net as pnet
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import constants as plugin_constants
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import excutils

from neutron.api import extensions as neutron_extensions
from neutron.db import api as db_api
from neutron.db import common_db_mixin
from neutron.db import extraroute_db
# Import l3_dvr_db to get the config options required for FWaaS
from neutron.db import l3_dvr_db  # noqa
from neutron.extensions import multiprovidernet as mpnet

from midonet.neutron.client import base as c_base
from midonet.neutron.common import config  # noqa
from midonet.neutron.common import constants as m_const
from midonet.neutron.db import l3_db_midonet
from midonet.neutron import extensions
from midonet.neutron.extensions import routerinterfacefip

LOG = logging.getLogger(__name__)


@registry.has_registry_receivers
class MidonetL3ServicePlugin(common_db_mixin.CommonDbMixin,
                             extraroute_db.ExtraRoute_db_mixin,
                             l3_db_midonet.MidonetL3DBMixin):

    """Implements L3 Router service plugin for Midonet."""

    supported_extension_aliases = ["router", "extraroute", "ext-gw-mode",
                                   "router-interface-fip", "fip64"]

    __native_pagination_support = True
    __native_sorting_support = True

    def __init__(self):
        super(MidonetL3ServicePlugin, self).__init__()

        # Instantiate MidoNet API client
        self.client = c_base.load_client(cfg.CONF.MIDONET)

        # Avoid any side effect from DVR getting set to true
        cfg.CONF.set_override("router_distributed", False)
        neutron_extensions.append_api_extensions_path(extensions.__path__)

    @classmethod
    def get_plugin_type(cls):
        return plugin_constants.L3

    def get_plugin_description(self):
        """Returns string description of the plugin."""
        return ("Midonet L3 Router Service Plugin")

    @staticmethod
    def _segments(network):
        if pnet.NETWORK_TYPE in network:
            yield {
                pnet.NETWORK_TYPE: network[pnet.NETWORK_TYPE],
            }
        segments = network.get(mpnet.SEGMENTS)
        if segments:
            for seg in segments:
                yield seg

    def _validate_network_type(self, context, network_id):
        our_types = [m_const.TYPE_MIDONET, m_const.TYPE_UPLINK]
        network = self._core_plugin.get_network(context, network_id)
        for seg in self._segments(network):
            if seg[pnet.NETWORK_TYPE] in our_types:
                return
        LOG.warning("Incompatible network %s", network)
        raise n_exc.BadRequest(resource='router', msg='Incompatible network')

    def _validate_router_gw_network(self, context, r):
        ext_gw_info = r.get(l3_apidef.EXTERNAL_GW_INFO)
        if ext_gw_info:
            self._validate_network_type(context, ext_gw_info['network_id'])

    @log_helpers.log_method_call
    @db_api.retry_if_session_inactive()
    def create_router(self, context, router):
        with db_api.context_manager.writer.using(context):
            # REVISIT(yamamoto): This should not call create_port inside
            # of a transaction.
            setattr(context, 'GUARD_TRANSACTION', False)
            r = super(MidonetL3ServicePlugin, self).create_router(context,
                                                                  router)
            self._validate_router_gw_network(context, r)
            self.client.create_router_precommit(context, r)

        try:
            self.client.create_router_postcommit(r)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create a router %(r_id)s in Midonet:"
                          "%(err)s", {"r_id": r["id"], "err": ex})
                try:
                    self.delete_router(context, r['id'])
                except Exception:
                    LOG.exception("Failed to delete a router %s", r["id"])
        return r

    @log_helpers.log_method_call
    @db_api.retry_if_session_inactive()
    def update_router(self, context, id, router):
        with db_api.context_manager.writer.using(context):
            # REVISIT(yamamoto): Updating external_gateway_info causes
            # create_port/delete_port.  This should not call them inside of
            # a transaction.
            setattr(context, 'GUARD_TRANSACTION', False)
            r = super(MidonetL3ServicePlugin, self).update_router(context, id,
                                                                  router)
            self._validate_router_gw_network(context, r)
            self.client.update_router_precommit(context, id, r)

        try:
            self.client.update_router_postcommit(id, r)
            if r['status'] != m_const.ROUTER_STATUS_ACTIVE:
                data = {'router': {'status': m_const.ROUTER_STATUS_ACTIVE}}
                r = super(MidonetL3ServicePlugin, self).update_router(
                    context, id, data)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to update a router %(r_id)s in MidoNet: "
                          "%(err)s", {"r_id": id, "err": ex})
                try:
                    data = {'router': {'status': m_const.ROUTER_STATUS_ERROR}}
                    super(MidonetL3ServicePlugin, self).update_router(
                        context, id, data)
                except Exception:
                    LOG.exception("Failed to update a router status %s", id)
        return r

    @log_helpers.log_method_call
    @db_api.retry_if_session_inactive()
    def delete_router(self, context, id):
        self._check_router_not_in_use(context, id)

        with db_api.context_manager.writer.using(context):
            # REVISIT(yamamoto): This should not call delete_port inside
            # of a transaction.
            setattr(context, 'GUARD_TRANSACTION', False)
            super(MidonetL3ServicePlugin, self).delete_router(context, id)
            self.client.delete_router_precommit(context, id)

        self.client.delete_router_postcommit(id)

    @log_helpers.log_method_call
    @db_api.retry_if_session_inactive()
    def add_router_interface(self, context, router_id, interface_info):
        by_port = bool(interface_info.get('port_id'))
        with db_api.context_manager.writer.using(context):
            # REVISIT(yamamoto): This should not call create_port/update_port
            # inside of a transaction.
            setattr(context, 'GUARD_TRANSACTION', False)
            info = super(MidonetL3ServicePlugin, self).add_router_interface(
                context, router_id, interface_info)
            self._validate_network_type(context, info['network_id'])
            self.client.add_router_interface_precommit(context, router_id,
                                                       info)

        try:
            self.client.add_router_interface_postcommit(router_id, info)
        except Exception as ex:
            LOG.error("Failed to create MidoNet resources to add router "
                      "interface. info=%(info)s, router_id=%(router_id)s, "
                      "error=%(err)r",
                      {"info": info, "router_id": router_id, "err": ex})
            with excutils.save_and_reraise_exception():
                if not by_port:
                    self.remove_router_interface(context, router_id, info)

        return info

    @log_helpers.log_method_call
    @db_api.retry_if_session_inactive()
    def remove_router_interface(self, context, router_id, interface_info):
        with db_api.context_manager.writer.using(context):
            # REVISIT(yamamoto): This should not call delete_port inside
            # of a transaction.
            setattr(context, 'GUARD_TRANSACTION', False)
            info = super(MidonetL3ServicePlugin, self).remove_router_interface(
                context, router_id, interface_info)
            self.client.remove_router_interface_precommit(context, router_id,
                                                          info)

        self.client.remove_router_interface_postcommit(router_id, info)
        return info

    @log_helpers.log_method_call
    @db_api.retry_if_session_inactive()
    def create_floatingip(self, context, floatingip):
        with db_api.context_manager.writer.using(context):
            # REVISIT(yamamoto): This should not call create_port inside
            # of a transaction.
            setattr(context, 'GUARD_TRANSACTION', False)
            fip = super(MidonetL3ServicePlugin, self).create_floatingip(
                context, floatingip)
            self.client.create_floatingip_precommit(context, fip)

        try:
            self.client.create_floatingip_postcommit(fip)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create floating ip %(fip)s: %(err)s",
                          {"fip": fip, "err": ex})
                try:
                    self.delete_floatingip(context, fip['id'])
                except Exception:
                    LOG.exception("Failed to delete a floating ip %s",
                                  fip['id'])
        return fip

    @log_helpers.log_method_call
    @db_api.retry_if_session_inactive()
    def delete_floatingip(self, context, id):
        with db_api.context_manager.writer.using(context):
            # REVISIT(yamamoto): This should not call delete_port inside
            # of a transaction.
            setattr(context, 'GUARD_TRANSACTION', False)
            super(MidonetL3ServicePlugin, self).delete_floatingip(context, id)
            self.client.delete_floatingip_precommit(context, id)

        self.client.delete_floatingip_postcommit(id)

    @log_helpers.log_method_call
    @db_api.retry_if_session_inactive()
    def update_floatingip(self, context, id, floatingip):
        with db_api.context_manager.writer.using(context):
            fip = super(MidonetL3ServicePlugin, self).update_floatingip(
                context, id, floatingip)
            self.client.update_floatingip_precommit(context, id, fip)

            # Update status based on association
            if fip.get('port_id') is None:
                new_status = n_const.FLOATINGIP_STATUS_DOWN
            else:
                new_status = n_const.FLOATINGIP_STATUS_ACTIVE
            self.update_floatingip_status(context, id, new_status)

        try:
            self.client.update_floatingip_postcommit(id, fip)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to update a floating ip "
                          "%(fip_id)s in MidoNet: %(err)s",
                          {"fip_id": id, "err": ex})
                try:
                    self.update_floatingip_status(
                        context, id, n_const.FLOATINGIP_STATUS_ERROR)
                except Exception:
                    LOG.exception("Failed to update floating ip status %s",
                                  id)
        return fip

    @registry.receives(resources.ROUTER_INTERFACE, [events.BEFORE_DELETE])
    def _check_router_interface_used_as_gw_for_fip(self, resource,
                                                   event, trigger, **kwargs):
        context = kwargs['context']
        router_id = kwargs['router_id']
        subnet_id = kwargs['subnet_id']
        if self._subnet_has_fip(context, router_id, subnet_id):
            raise routerinterfacefip.RouterInterfaceInUseAsGatewayByFloatingIP(
                router_id=router_id, subnet_id=subnet_id)



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import extensions
from neutron_lib import exceptions as nexception

from midonet.neutron._i18n import _
from midonet.neutron.common import constants as m_const


class MidonetBgpPeerInUse(nexception.InUse):
    message = _("bgp peer %(id)s %(reason)s")

    def __init__(self, **kwargs):
        if 'reason' not in kwargs:
            kwargs['reason'] = "is already associated with bgp speaker"
        super(MidonetBgpPeerInUse, self).__init__(**kwargs)


class NetworkTypeInvalid(nexception.InvalidInput):
    message = _("Only external network can be specified.")


class ExternalNetworkUnbound(nexception.BadRequest):
    message = _("Unable to complete operation for bgp speaker. "
                "External network must be associated with bgp speaker when "
                "logical_router is not specified in bgp speaker creation.")


class BgpSpeakerInUse(nexception.InUse):
    message = _("Bgp speaker %(id)s %(reason)s")

    def __init__(self, **kwargs):
        if 'reason' not in kwargs:
            kwargs['reason'] = "is still associated with bgp peers"
        super(BgpSpeakerInUse, self).__init__(**kwargs)


class NoSubnetInNetwork(nexception.InvalidInput):
    message = _("No subnets in the network: %(network_id)s.")


class NoGatewayIpOnSubnet(nexception.InvalidInput):
    message = _("No gateway ips on the subnet: %(subnet_id)s.")


class NoGatewayIpPortOnSubnet(nexception.InvalidInput):
    message = _("No ports have gateway ip on the subnet: %(subnet_id)s.")


EXTENDED_ATTRIBUTES_2_0 = {
    'bgp-speakers': {
        m_const.LOGICAL_ROUTER: {'allow_post': True, 'allow_put': False,
                                 'validate': {'type:uuid_or_none': None},
                                 'is_visible': True, 'default': None},
    }
}

BGP_ROUTER_EXT_ALIAS = "bgp-speaker-router-insertion"


class Bgp_speaker_router_insertion(extensions.ExtensionDescriptor):
    """Extension class supporting BgpSpeaker and Router association.

    """
    @classmethod
    def get_name(cls):
        return "BgpSpeakerRouterInsertion"

    @classmethod
    def get_alias(cls):
        return BGP_ROUTER_EXT_ALIAS

    @classmethod
    def get_description(cls):
        return "Bgp Speaker Router insertion on specified router"

    @classmethod
    def get_updated(cls):
        return "2016-04-17T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}

    def get_required_extensions(self):
        return ["bgp"]



# Copyright (c) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import testtools

from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
import tempest.lib.exceptions as lib_exc

from neutron_tempest_plugin.api import base


class ExpectedException(testtools.ExpectedException):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if super(ExpectedException, self).__exit__(exc_type, exc_value, tb):
            self.exception = exc_value
            return True
        return False


class RouterInterfaceFip(base.BaseAdminNetworkTest):
    @classmethod
    @utils.requires_ext(extension="router-interface-fip", service="network")
    def resource_setup(cls):
        super(RouterInterfaceFip, cls).resource_setup()

    @decorators.idempotent_id('943ab44d-0ea7-4c6a-bdfd-8ba759622992')
    def test_router_interface_fip(self):
        # +-------------+
        # | router1     |
        # +-+--------+--+
        #   |        |
        # +-+--+   +-+--------+
        # |net1|   |net2      |
        # |    |   |(external)|
        # +-+--+   +--+-------+
        #   |         |
        #  port1     fip2
        cidr1 = netaddr.IPNetwork('192.2.1.0/24')
        cidr2 = netaddr.IPNetwork('192.2.2.0/24')
        router1_name = data_utils.rand_name('router1')
        router1 = self.create_router(router1_name)
        net1 = self.create_network()
        subnet1 = self.create_subnet(net1, cidr=cidr1)
        self.create_router_interface(router1['id'], subnet1['id'])
        net2 = self.admin_client.create_network(
            project_id=self.client.tenant_id,
            **{'router:external': True})['network']
        self.networks.append(net2)
        subnet2 = self.create_subnet(net2, cidr=cidr2)
        self.create_router_interface(router1['id'], subnet2['id'])
        port1 = self.create_port(net1)
        fip2 = self.create_floatingip(net2['id'])
        fip2_updated = self.client.update_floatingip(
            fip2['id'], port_id=port1['id'])['floatingip']
        expected = {
            'floating_network_id': net2['id'],
            'port_id': port1['id'],
            'router_id': router1['id'],
        }
        for k, v in expected.items():
            self.assertIn(k, fip2_updated)
            self.assertEqual(v, fip2_updated[k])
        if 'revision_number' in fip2:
            self.assertGreater(fip2_updated['revision_number'],
                               fip2['revision_number'])
        # NOTE(yamamoto): The status can be updated asynchronously.
        fip2_shown = self.client.show_floatingip(fip2['id'])['floatingip']
        if 'revision_number' in fip2:
            self.assertGreaterEqual(fip2_shown['revision_number'],
                                    fip2_updated['revision_number'])
        fip2_shown.pop('status')
        fip2_shown.pop('updated_at')
        fip2_shown.pop('revision_number')
        fip2_updated.pop('status')
        fip2_updated.pop('updated_at')
        fip2_updated.pop('revision_number')
        self.assertEqual(fip2_updated, fip2_shown)
        with ExpectedException(lib_exc.Conflict) as ctx:
            self.client.remove_router_interface_with_subnet_id(
                router1['id'], subnet2['id'])
        self.assertEqual('RouterInterfaceInUseAsGatewayByFloatingIP',
                         ctx.exception.resp_body['type'])
        with ExpectedException(lib_exc.Conflict) as ctx:
            self.client.remove_router_interface_with_subnet_id(
                router1['id'], subnet1['id'])
        self.assertEqual('RouterInterfaceInUseByFloatingIP',
                         ctx.exception.resp_body['type'])
        fip2_updated2 = self.client.update_floatingip(
            fip2['id'], port_id=None)['floatingip']
        expected = {
            'floating_network_id': net2['id'],
            'floating_ip_address': fip2_shown['floating_ip_address'],
            'port_id': None,
            'router_id': None,
        }
        for k, v in expected.items():
            self.assertIn(k, fip2_updated2)
            self.assertEqual(v, fip2_updated2[k])
        self.client.remove_router_interface_with_subnet_id(
            router1['id'], subnet2['id'])
        self.client.remove_router_interface_with_subnet_id(
            router1['id'], subnet1['id'])



# Copyright (C) 2015 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import excutils

from neutron.api import extensions as neutron_extensions
from neutron_fwaas.db.firewall import firewall_db
from neutron_fwaas import extensions
from neutron_fwaas.services.firewall import fwaas_plugin as fw_plugin

from midonet.neutron.client import base as c_base
from midonet.neutron.common import config  # noqa

LOG = logging.getLogger(__name__)


# TODO(yamamoto): This driver needs a major restructure for task-based api.
# I'm not sure if it's worth the effort given that FWaaS v2 things are
# coming.  FWaaS v2 will have a better driver api.


class _MidonetFirewallDriver(object):
    """FWaaS driver for MidoNet that implements the RPC API

    The driver does not actually do any RPC, but simply proxies the API calls
    to the MidoNet side.  This design was chosen instead of implementing the
    FWaaS plugin methods because these methods get as input more complete data
    on the firewall object, including its rules and the routers
    associated/disassociated.  The downside is that each request might get
    bloated unnecessarily.  We will revisit the design when such problem
    arises.
    """

    def __init__(self, client, callbacks):
        self.client = client
        self.callbacks = callbacks
        self.plugin = callbacks.plugin

    @log_helpers.log_method_call
    def create_firewall(self, context, firewall, host=None):
        # This method is called outside of DB transaction
        try:
            self.client.create_firewall(context, firewall)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create a firewall %(fw_id)s "
                          "in Midonet: %(err)s",
                          {"fw_id": firewall["id"], "err": ex})
                try:
                    self.plugin.delete_db_firewall_object(context,
                                                          firewall['id'])
                except Exception:
                    LOG.exception("Failed to delete firewall %s",
                                  firewall['id'])

        self._set_firewall_status_noerror(context, firewall)

    @log_helpers.log_method_call
    def update_firewall(self, context, firewall, host=None):
        # This method is called outside of DB transaction
        try:
            self.client.update_firewall(context, firewall)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to update a firewall %(fw_id)s "
                          "in Midonet: %(err)s",
                          {"fw_id": firewall["id"], "err": ex})
                try:
                    self.callbacks.set_firewall_status(context, firewall['id'],
                                                       constants.ERROR)
                except Exception:
                    LOG.exception("Failed to update firewall status %s",
                                  firewall['id'])

        self._set_firewall_status_noerror(context, firewall)

    def _set_firewall_status_noerror(self, context, firewall):
        if firewall['add-router-ids']:
            status = constants.ACTIVE
        else:
            status = constants.INACTIVE
        self.callbacks.set_firewall_status(context, firewall['id'], status)

    @log_helpers.log_method_call
    def delete_firewall(self, context, firewall, host=None):
        # This method is called outside of DB transaction
        try:
            self.client.delete_firewall(context, firewall)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to delete a firewall %(fw_id)s "
                          "in Midonet: %(err)s",
                          {"fw_id": firewall["id"], "err": ex})
                try:
                    self.callbacks.set_firewall_status(context, firewall['id'],
                                                       constants.ERROR)
                except Exception:
                    LOG.exception("Failed to update firewall status %s",
                                  firewall['id'])

        self.callbacks.firewall_deleted(context, firewall['id'])


class MidonetFirewallPlugin(fw_plugin.FirewallPlugin):

    """Implementation of the Neutron Firewall Service Plugin.

    This class manages the workflow of FWaaS request/response.
    DB related updates, including router insertion logics, are handled by
    fw_plugin.FirewallPlugin.
    """

    def __init__(self):
        # Override initialization to avoid any RPC setup as MidoNet does not
        # rely on any agent to implement FWaaS.  Instead, set the rpc handling
        # to the _MidonetFirewallDriver class so that it handles the FWaaS
        # update events.

        # Register the FWaaS extensions path
        neutron_extensions.append_api_extensions_path(extensions.__path__)

        # Although callbacks are unnecessary in midonet, use FirewallCallbacks
        # because it contains useful methods for DB updates.
        self.callbacks = fw_plugin.FirewallCallbacks(self)
        self.client = c_base.load_client(cfg.CONF.MIDONET)
        self.agent_rpc = _MidonetFirewallDriver(self.client, self.callbacks)
        self.endpoints = [self.callbacks]  # So that tests don't complain
        # TODO(yamamoto): Remove this subscribe() call once neutron-fwaas
        # was converted to use registry decorators.
        firewall_db.subscribe()



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import validators
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib.db import model_base
from neutron_lib import exceptions as n_exc
from neutron_lib.exceptions import l3 as l3_exc
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_db import exception as db_exc
from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy.ext import declarative
from sqlalchemy import orm
from sqlalchemy.orm import exc

from neutron.db import common_db_mixin

from midonet.neutron.extensions import gateway_device as gw_device_ext

GATEWAY_DEVICES = 'midonet_gateway_devices'
GATEWAY_HW_VTEP_DEVICES = 'midonet_gateway_hw_vtep_devices'
GATEWAY_OVERLAY_ROUTER_DEVICES = 'midonet_gateway_overlay_router_devices'
GATEWAY_NETWORK_VLAN_DEVICES = 'midonet_gateway_network_vlan_devices'
GATEWAY_TUNNEL_IPS = 'midonet_gateway_tunnel_ips'
GATEWAY_REMOTE_MAC_TABLES = 'midonet_gateway_remote_mac_tables'


class GatewayDevice(model_base.BASEV2, model_base.HasProjectNoIndex):
    """Represents a gateway device."""

    __tablename__ = GATEWAY_DEVICES
    id = sa.Column(sa.String(36), primary_key=True)
    name = sa.Column(sa.String(255))
    type = sa.Column(sa.String(length=255), nullable=False)


class GatewayHwVtepDevice(model_base.BASEV2):
    """Represents a gateway hardware vtep device."""

    __tablename__ = GATEWAY_HW_VTEP_DEVICES
    device_id = sa.Column(sa.String(36),
                          sa.ForeignKey('midonet_gateway_devices.id',
                          ondelete="CASCADE"),
                          nullable=False, primary_key=True)
    management_ip = sa.Column(sa.String(length=64), nullable=False,
                              unique=True)
    management_port = sa.Column(sa.Integer(), nullable=False)
    management_protocol = sa.Column(sa.String(length=255),
                                    nullable=False)
    gateway_device = orm.relationship(
        GatewayDevice,
        backref=orm.backref('hw_vtep', cascade='delete', lazy='joined'),
        primaryjoin="GatewayDevice.id==GatewayHwVtepDevice.device_id")


class GatewayVirtualDevice(object):
    """Represents a virtual gateway device."""
    device_id = sa.Column(sa.String(36), nullable=False, primary_key=True)

    @declarative.declared_attr
    def __table_args__(cls):
        return (
            sa.ForeignKeyConstraint(['device_id'],
                                    ['midonet_gateway_devices.id'],
                                    ondelete="CASCADE"),
            model_base.BASEV2.__table_args__,
        )


def get_type_model_map():
    return {table.resource_type: table
            for table in GatewayVirtualDevice.__subclasses__()}


def _resource_id_column(foreign_key):
    return sa.Column(sa.String(36),
                     sa.ForeignKey(foreign_key),
                     nullable=False)


def _gateway_device_relation(class_name, ref_key):
    relation = "GatewayDevice.id==" + class_name + ".device_id"
    return orm.relationship(
        GatewayDevice,
        backref=orm.backref(ref_key, cascade='delete', lazy='joined'),
        primaryjoin=relation)


class GatewayOverlayRouterDevice(GatewayVirtualDevice, model_base.BASEV2):
    """Represents a gateway overlay router device."""

    __tablename__ = GATEWAY_OVERLAY_ROUTER_DEVICES
    resource_id = _resource_id_column('routers.id')
    gateway_device = _gateway_device_relation('GatewayOverlayRouterDevice',
                                              'overlay_router')
    resource_type = gw_device_ext.ROUTER_DEVICE_TYPE


class GatewayVlanNetworkDevice(GatewayVirtualDevice, model_base.BASEV2):
    """Represents a gateway vlan network device."""

    __tablename__ = GATEWAY_NETWORK_VLAN_DEVICES
    resource_id = _resource_id_column('networks.id')
    gateway_device = _gateway_device_relation('GatewayVlanNetworkDevice',
                                              'vlan_network')
    resource_type = gw_device_ext.NETWORK_VLAN_TYPE


class GatewayTunnelIp(model_base.BASEV2):
    """Represents a IP address for VTEP."""

    __tablename__ = GATEWAY_TUNNEL_IPS
    device_id = sa.Column(sa.String(36),
                          sa.ForeignKey('midonet_gateway_devices.id',
                          ondelete="CASCADE"),
                          nullable=False, primary_key=True)
    tunnel_ip = sa.Column(sa.String(64), nullable=False, unique=True)
    gateway_device = orm.relationship(
        GatewayDevice,
        backref=orm.backref('tunnel_ip_list', cascade='delete', lazy='joined'),
        primaryjoin="GatewayDevice.id==GatewayTunnelIp.device_id")


class GatewayRemoteMacTable(model_base.BASEV2):
    """Represents a mac table for vtep."""

    __tablename__ = GATEWAY_REMOTE_MAC_TABLES
    id = sa.Column(sa.String(36), primary_key=True)
    device_id = sa.Column(sa.String(36),
                          sa.ForeignKey('midonet_gateway_devices.id',
                          ondelete="CASCADE"),
                          nullable=False)
    mac_address = sa.Column(sa.String(length=32), nullable=False,
                            unique=True)
    vtep_address = sa.Column(sa.String(length=64), nullable=False)
    segmentation_id = sa.Column(sa.Integer())
    gateway_device = orm.relationship(
        GatewayDevice,
        backref=orm.backref('mac_table_list', cascade='delete', lazy='joined'),
        primaryjoin="GatewayDevice.id==GatewayRemoteMacTable.device_id")


@registry.has_registry_receivers
class GwDeviceDbMixin(gw_device_ext.GwDevicePluginBase,
                      common_db_mixin.CommonDbMixin):
    """Mixin class to add gateway device."""

    __native_bulk_support = False

    def _ensure_gateway_device_not_in_use(self, context, gw_dev_id):
        """Ensure that gateway_device is not in use."""

        l2gw_plugin = directory.get_plugin('L2GW')
        if l2gw_plugin:
            if l2gw_plugin._get_l2gw_devices_by_device_id(context, gw_dev_id):
                raise gw_device_ext.GatewayDeviceInUse(id=gw_dev_id)
        gw_dev = self._get_gateway_device(context, gw_dev_id)

        return gw_dev

    def _get_gateway_device(self, context, id):
        try:
            query = self._model_query(context, GatewayDevice)
            gw_dev_db = query.filter(GatewayDevice.id == id).one()

        except exc.NoResultFound:
            raise gw_device_ext.GatewayDeviceNotFound(id=id)

        return gw_dev_db

    def _get_gateway_device_from_resource(self, context, resource_type,
                                          resource_id):
        try:
            device_model = get_type_model_map()[resource_type]
            query = self._model_query(context, device_model)
            gw_dev_db = query.filter(
                device_model.resource_id == resource_id).one()

        except exc.NoResultFound:
            pass

        else:
            return gw_dev_db

    def _get_hw_vtep_from_management_ip(self, context, management_ip):
        try:
            query = self._model_query(context, GatewayHwVtepDevice)
            gw_hw_vtep_db = query.filter(
                GatewayHwVtepDevice.management_ip == management_ip).one()

        except exc.NoResultFound:
            pass

        else:
            return gw_hw_vtep_db

    def _get_remote_mac_entry(self, context, id, gateway_device_id):
        try:
            query = self._model_query(context, GatewayRemoteMacTable)
            rmt_db = query.filter(GatewayRemoteMacTable.id == id).one()
            if rmt_db.device_id != gateway_device_id:
                raise gw_device_ext.RemoteMacEntryWrongGatewayDevice(
                    id=id, gateway_device_id=gateway_device_id)

        except exc.NoResultFound:
            raise gw_device_ext.RemoteMacEntryNotFound(id=id)

        return rmt_db

    def _get_tunnel_ip_from_ip_address(self, context, ip):
        try:
            query = self._model_query(context, GatewayTunnelIp)
            gw_tun_ip_db = query.filter(
                GatewayTunnelIp.tunnel_ip == ip).one()

        except exc.NoResultFound:
            pass

        else:
            return gw_tun_ip_db

    def _make_gateway_device_dict(self, gw_dev_db, fields=None):
        res = {'id': gw_dev_db['id'],
               'name': gw_dev_db['name'],
               'type': gw_dev_db['type'],
               'tenant_id': gw_dev_db['tenant_id'],
               'tunnel_ips': [],
               'remote_mac_entries': []}
        if gw_dev_db['type'] == gw_device_ext.NETWORK_VLAN_TYPE:
            res['management_ip'] = None
            res['management_port'] = None
            res['management_protocol'] = None
            res['resource_id'] = gw_dev_db.vlan_network[0]['resource_id']
            # tunnel_ips and remote_mac_entries are not set
            return self._fields(res, fields)

        if gw_dev_db['type'] == gw_device_ext.HW_VTEP_TYPE:
            hw_vtep = gw_dev_db.hw_vtep[0]
            res['management_ip'] = hw_vtep['management_ip']
            res['management_port'] = hw_vtep['management_port']
            res['management_protocol'] = hw_vtep['management_protocol']
            res['resource_id'] = ""
        if gw_dev_db['type'] == gw_device_ext.ROUTER_DEVICE_TYPE:
            res['management_ip'] = None
            res['management_port'] = None
            res['management_protocol'] = None
            res['resource_id'] = gw_dev_db.overlay_router[0]['resource_id']
        res['tunnel_ips'] = list(map(
            lambda n: n['tunnel_ip'], gw_dev_db.tunnel_ip_list))
        for item in gw_dev_db.mac_table_list:
            entry = {'id': item['id'],
                     'mac_address': item['mac_address'],
                     'vtep_address': item['vtep_address'],
                     'segmentation_id': item['segmentation_id']}
            res['remote_mac_entries'].append(entry)

        return self._fields(res, fields)

    def _make_remote_mac_dict(self, gw_rme_db, fields=None):
        res = {'id': gw_rme_db['id'],
               'mac_address': gw_rme_db['mac_address'],
               'vtep_address': gw_rme_db['vtep_address'],
               'segmentation_id': gw_rme_db['segmentation_id'],
               'device_id': gw_rme_db['device_id']}
        return self._fields(res, fields)

    def _tunnel_ip_db_add(self, context, gw_dev_id, add_ips):
        for ip in add_ips:
            tun_ip_db = GatewayTunnelIp(
                device_id=gw_dev_id,
                tunnel_ip=ip)
            context.session.add(tun_ip_db)

    def _tunnel_ip_db_delete(self, context, gw_dev_id, delete_ips):
        for ip in delete_ips:
            tun_ip_db = self._get_tunnel_ip_from_ip_address(context, ip)
            context.session.delete(tun_ip_db)

    def _update_gateway_device_db(self, context, gw_dev_id, data):
        """Update the DB object.

        following parameter can be updated.
            name, tunnel_ips
        """

        add_ips, delete_ips = [], []
        with context.session.begin(subtransactions=True):
            gw_dev_db = self._get_gateway_device(context, gw_dev_id)
            if data.get('tunnel_ips'):
                exist_ips = list(map(
                    lambda n: n['tunnel_ip'], gw_dev_db.tunnel_ip_list))
                add_ips = set(data['tunnel_ips']) - set(exist_ips)
                delete_ips = set(exist_ips) - set(data['tunnel_ips'])
            if delete_ips:
                self._tunnel_ip_db_delete(context, gw_dev_id, delete_ips)
            if add_ips:
                self._tunnel_ip_db_add(context, gw_dev_id, add_ips)
            if data:
                gw_dev_db.update(data)
        context.session.refresh(gw_dev_db)

        return gw_dev_db

    def _validate_gateway_device(self, context, gw_dev):
        if gw_dev['type'] == gw_device_ext.HW_VTEP_TYPE:
            self._validate_hw_vtep(gw_dev, context)
        if gw_dev['type'] == gw_device_ext.ROUTER_DEVICE_TYPE:
            self._validate_router_vtep(gw_dev, context)
        if gw_dev['type'] == gw_device_ext.NETWORK_VLAN_TYPE:
            self._validate_vlan_network(gw_dev, context)

    def _validate_hw_vtep(self, gw_dev, context):
        if not gw_dev['management_ip'] or not gw_dev['management_port']:
            raise gw_device_ext.HwVtepTypeInvalid(type=gw_dev['type'])
        if self._get_hw_vtep_from_management_ip(
                context, gw_dev['management_ip']):
            raise gw_device_ext.GatewayDeviceParamDuplicate(
                param_name='management_ip',
                param_value=gw_dev['management_ip'])

    def _validate_resource_router_vtep(self, context, router_id):
        # Check specified router existence
        l3plugin = directory.get_plugin(constants.L3)

        try:
            l3plugin.get_router(context, router_id)
        except l3_exc.RouterNotFound:
            raise gw_device_ext.ResourceNotFound(resource_id=router_id)

        if self._get_gateway_device_from_resource(
                context, gw_device_ext.ROUTER_DEVICE_TYPE, router_id):
            raise gw_device_ext.DeviceInUseByGatewayDevice(
                resource_id=router_id, resource_type='router')

    def _validate_router_vtep(self, gw_dev, context):
        if not gw_dev['resource_id']:
            raise gw_device_ext.RouterVtepTypeInvalid(type=gw_dev['type'])
        self._validate_resource_router_vtep(context,
                                            gw_dev['resource_id'])

    def _validate_resource_vlan_network(self, context, network_id):
        # Check specified netowrk existence
        core_plugin = directory.get_plugin()

        try:
            core_plugin.get_network(context, network_id)
        except n_exc.NetworkNotFound:
            raise gw_device_ext.ResourceNotFound(resource_id=network_id)

        if self._get_gateway_device_from_resource(
                context, gw_device_ext.NETWORK_VLAN_TYPE, network_id):
            raise gw_device_ext.DeviceInUseByGatewayDevice(
                resource_id=network_id, resource_type='network')

    def _validate_vlan_network(self, gw_dev, context):
        if not gw_dev['resource_id']:
            raise gw_device_ext.NetworkVlanTypeInvalid(type=gw_dev['type'])
        self._validate_resource_vlan_network(context,
                                             gw_dev['resource_id'])

    def _validate_tunnel_ips(self, tunnel_ips, gw_type):
        if len(tunnel_ips) > 1:
            raise gw_device_ext.TunnelIPsExhausted()
        if len(tunnel_ips) == 0:
            raise gw_device_ext.TunnelIPsRequired(gw_type=gw_type)

    def create_gateway_device(self, context, gw_device):
        """Create a gateway device"""
        gw_dev = gw_device['gateway_device']
        tenant_id = gw_dev['tenant_id']
        self._validate_gateway_device(context, gw_dev)
        if gw_dev['type'] != gw_device_ext.NETWORK_VLAN_TYPE:
            self._validate_tunnel_ips(gw_dev.get('tunnel_ips') or [],
                                      gw_dev['type'])

        with context.session.begin(subtransactions=True):
            gw_dev_db = GatewayDevice(
                id=uuidutils.generate_uuid(),
                name=gw_dev['name'],
                type=(gw_dev['type'] or gw_device_ext.HW_VTEP_TYPE),
                tenant_id=tenant_id)
            context.session.add(gw_dev_db)
            if gw_dev_db['type'] == gw_device_ext.HW_VTEP_TYPE:
                gw_hw_vtep_db = GatewayHwVtepDevice(
                    device_id=gw_dev_db['id'],
                    management_ip=gw_dev['management_ip'],
                    management_port=gw_dev['management_port'],
                    management_protocol=(
                        gw_dev['management_protocol'] or gw_device_ext.OVSDB))
                context.session.add(gw_hw_vtep_db)

            device_model = get_type_model_map().get(gw_dev_db['type'])
            if device_model:
                gw_vlan_db = device_model(
                    device_id=gw_dev_db['id'],
                    resource_id=gw_dev['resource_id'])
                context.session.add(gw_vlan_db)

            # In network vlan type gateway device, tunnel_ips should be None
            if gw_dev.get('tunnel_ips'):
                self._tunnel_ip_db_add(context,
                                       gw_dev_db['id'], gw_dev['tunnel_ips'])

        return self._make_gateway_device_dict(gw_dev_db)

    def create_gateway_device_remote_mac_entry(self, context,
                                               gateway_device_id,
                                               remote_mac_entry):
        rme = remote_mac_entry['remote_mac_entry']

        try:
            with context.session.begin(subtransactions=True):
                gw_rmt_db = GatewayRemoteMacTable(
                    id=uuidutils.generate_uuid(),
                    device_id=gateway_device_id,
                    mac_address=rme['mac_address'],
                    segmentation_id=rme['segmentation_id'],
                    vtep_address=rme['vtep_address'])
                context.session.add(gw_rmt_db)
        except db_exc.DBDuplicateEntry:
            raise gw_device_ext.DuplicateRemoteMacEntry(
                mac_address=rme['mac_address'])

        return self._make_remote_mac_dict(gw_rmt_db)

    def delete_gateway_device(self, context, id):
        gw_dev = self._ensure_gateway_device_not_in_use(context, id)
        with context.session.begin(subtransactions=True):
            context.session.delete(gw_dev)

    def delete_gateway_device_remote_mac_entry(self, context, id,
                                               gateway_device_id):
        rmt_db = self._get_remote_mac_entry(context, id, gateway_device_id)
        with context.session.begin(subtransactions=True):
            context.session.delete(rmt_db)

    def get_gateway_device(self, context, id, fields=None):
        gw_dev = self._get_gateway_device(context, id)
        return self._make_gateway_device_dict(gw_dev, fields)

    def get_gateway_devices(self, context, filters=None, fields=None,
                            sorts=None, limit=None, marker=None,
                            page_reverse=False):
        marker_obj = self._get_marker_obj(context, 'gateway_device', limit,
                                          marker)

        return self._get_collection(context,
                                    GatewayDevice,
                                    self._make_gateway_device_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts,
                                    limit=limit, marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    def get_gateway_device_remote_mac_entries(self, context, gateway_device_id,
                                              filters=None, fields=None,
                                              sorts=None, limit=None,
                                              marker=None, page_reverse=False):
        marker_obj = self._get_marker_obj(context, 'remote_mac_entry', limit,
                                          marker)
        filters['device_id'] = [gateway_device_id]
        return self._get_collection(context,
                                    GatewayRemoteMacTable,
                                    self._make_remote_mac_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts,
                                    limit=limit, marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    def get_gateway_device_remote_mac_entry(self, context, id,
                                            gateway_device_id, fields=None):
        rme = self._get_remote_mac_entry(context, id, gateway_device_id)
        return self._make_remote_mac_dict(rme, fields)

    def update_gateway_device(self, context, id, gw_device):
        gw_dev = gw_device['gateway_device']
        gw_dev_db = self._get_gateway_device(context, id)
        if gw_dev_db.type == gw_device_ext.NETWORK_VLAN_TYPE:
            del gw_device['gateway_device']['tunnel_ips']
        elif validators.is_attr_set(gw_dev.get('tunnel_ips')):
            self._validate_tunnel_ips(gw_dev.get('tunnel_ips'),
                                      gw_dev_db.type)
        gw_dev_db = self._update_gateway_device_db(context, id, gw_dev)
        return self._make_gateway_device_dict(gw_dev_db)

    @registry.receives(resources.ROUTER, [events.BEFORE_DELETE])
    @registry.receives(resources.NETWORK, [events.PRECOMMIT_DELETE])
    def _gateway_device_callback(self, resource, event, trigger, **kwargs):
        if resource == resources.ROUTER:
            resource_id = kwargs['router_id']
            gw_dev_type = gw_device_ext.ROUTER_DEVICE_TYPE
            resource_type = 'router'
        elif resource == resources.NETWORK:
            resource_id = kwargs['network_id']
            gw_dev_type = gw_device_ext.NETWORK_VLAN_TYPE
            resource_type = 'network'
        else:
            return
        context = kwargs.get('context')
        if self._get_gateway_device_from_resource(context,
                                                  gw_dev_type,
                                                  resource_id):
            raise gw_device_ext.DeviceInUseByGatewayDevice(
                resource_id=resource_id, resource_type=resource_type)



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from neutronclient import shell

from midonet.neutron.extensions import logging_resource as log_res_ext
from midonet.neutron.tests.unit.neutronclient_ext import test_cli20
from midonet.neutron.tests.unit import test_extension_logging_resource as telr
from midonet.neutronclient.logging_resource_extension import _firewall_log

RESOURCE = 'firewall_log'
RESOURCES = 'firewall_logs'
FAKE_LOGGING_RESOURCE_ID = 'my_logging_resource_id'
FAKE_FIREWALL_ID = 'my_firewall_id'
FAKE_FIREWALL_LOG_ID = 'my_firewall_log_id'


class CLITestV20FirewallLogJSON(test_cli20.CLIExtTestV20Base):

    def setUp(self):
        firewall_log = ("firewall_log", _firewall_log)
        self._mock_load_extensions(firewall_log)
        super(CLITestV20FirewallLogJSON,
              self).setUp(plurals={'firewall_logs': 'firewall_log'})
        self.register_non_admin_status_resource('firewall_log')

    def test_firewall_log_cmd_loaded(self):
        neutron_shell = shell.NeutronShell('2.0')
        firewall_log_cmd = {'logging-firewall-list':
                            _firewall_log.FirewallLogList,
                            'logging-firewall-create':
                            _firewall_log.FirewallLogCreate,
                            'logging-firewall-update':
                            _firewall_log.FirewallLogUpdate,
                            'logging-firewall-delete':
                            _firewall_log.FirewallLogDelete,
                            'logging-firewall-show':
                            _firewall_log.FirewallLogShow}
        for cmd_name, cmd_class in firewall_log_cmd.items():
            found = neutron_shell.command_manager.find_command([cmd_name])
            self.assertEqual(cmd_class, found[0])

    def _create_firewall_log(self, args, position_names,
                             position_values, parent_id=None):
        cmd = _firewall_log.FirewallLogCreate(
            test_cli20.MyApp(sys.stdout), None)
        self._test_create_resource(RESOURCE, cmd, None, FAKE_FIREWALL_LOG_ID,
                                   args, position_names, position_values,
                                   parent_id=parent_id)

    def _update_firewall_log(self, args, values, parent_id=None):
        cmd = _firewall_log.FirewallLogUpdate(
            test_cli20.MyApp(sys.stdout), None)
        self._test_update_ext_resource(RESOURCE, cmd, FAKE_FIREWALL_LOG_ID,
                                       args, values, parent_id=parent_id)

    def test_create_firewall_log(self):
        args = ['--description', telr.FAKE_FW_LOG_DESC,
                '--fw-event', log_res_ext.FW_EVENT_DROP,
                '--firewall-id', FAKE_FIREWALL_ID,
                FAKE_LOGGING_RESOURCE_ID]
        position_names = ['description', 'fw_event', 'firewall_id']
        position_values = [telr.FAKE_FW_LOG_DESC, log_res_ext.FW_EVENT_DROP,
                           FAKE_FIREWALL_ID]
        self._create_firewall_log(args, position_names,
                                  position_values,
                                  parent_id=FAKE_LOGGING_RESOURCE_ID)

    def test_create_firewall_log_with_missing_logging_resource_id(self):
        args = ['--description', telr.FAKE_FW_LOG_DESC,
                '--fw-event', log_res_ext.FW_EVENT_DROP,
                '--firewall-id', FAKE_FIREWALL_ID]
        position_names = []
        position_values = []
        self.assertRaises(SystemExit, self._create_firewall_log,
                          args, position_names, position_values)

    def test_create_firewall_log_with_missing_firewall_id(self):
        args = ['--description', telr.FAKE_FW_LOG_DESC,
                '--fw-event', log_res_ext.FW_EVENT_DROP,
                FAKE_LOGGING_RESOURCE_ID]
        position_names = []
        position_values = []
        self.assertRaises(SystemExit, self._create_firewall_log,
                          args, position_names, position_values,
                          parent_id=FAKE_LOGGING_RESOURCE_ID)

    def test_update_firewall_log_with_description(self):
        args = [FAKE_FIREWALL_LOG_ID, FAKE_LOGGING_RESOURCE_ID,
                '--description', telr.NEW_FW_LOG_DESC]
        values = {'description': telr.NEW_FW_LOG_DESC}
        self._update_firewall_log(args, values,
                                  parent_id=FAKE_LOGGING_RESOURCE_ID)

    def test_update_firewall_log_with_fw_event(self):
        args = [FAKE_FIREWALL_LOG_ID, FAKE_LOGGING_RESOURCE_ID,
                '--fw-event', log_res_ext.FW_EVENT_ACCEPT]
        values = {'fw_event': log_res_ext.FW_EVENT_ACCEPT}
        self._update_firewall_log(args, values,
                                  parent_id=FAKE_LOGGING_RESOURCE_ID)

    def test_delete_firewall_log(self):
        cmd = _firewall_log.FirewallLogDelete(
            test_cli20.MyApp(sys.stdout), None)
        args = [FAKE_FIREWALL_LOG_ID, FAKE_LOGGING_RESOURCE_ID]
        self._test_delete_ext_resource(RESOURCE, cmd,
                                       FAKE_FIREWALL_LOG_ID, args,
                                       parent_id=FAKE_LOGGING_RESOURCE_ID)

    def test_list_firewall_logs(self):
        cmd = _firewall_log.FirewallLogList(
            test_cli20.MyApp(sys.stdout), None)
        args = [FAKE_LOGGING_RESOURCE_ID]
        self._test_list_resources(RESOURCES, cmd, base_args=args,
                                  parent_id=FAKE_LOGGING_RESOURCE_ID)

    def test_show_firewall_log(self):
        cmd = _firewall_log.FirewallLogShow(
            test_cli20.MyApp(sys.stdout), None)
        args = [FAKE_FIREWALL_LOG_ID, FAKE_LOGGING_RESOURCE_ID]
        self._test_show_ext_resource(RESOURCE, cmd,
                                     FAKE_FIREWALL_LOG_ID, args,
                                     parent_id=FAKE_LOGGING_RESOURCE_ID)



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib.exceptions import l3 as l3_exc
from neutron_lib.plugins import directory
from oslo_db import exception as db_exc
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from sqlalchemy.orm import exc

from neutron.db import api as db_api
from neutron_dynamic_routing.db import bgp_db as bdb

from midonet.neutron.db import bgp_speaker_router_insertion_model as model
from midonet.neutron.extensions import bgp_speaker_router_insertion as bsri


LOG = logging.getLogger(__name__)


@registry.has_registry_receivers
class BgpSpeakerRouterInsertionDbMixin(object):

    """Access methods for the bgp_speaker_router_associations table."""

    @log_helpers.log_method_call
    def set_router_for_bgp_speaker(self, context, bgp_sp_id, r_id):
        """Sets the routers associated with the bgp speaker."""
        try:
            with db_api.context_manager.writer.using(context):
                bgp_router_db = model.BgpSpeakerRouterAssociation(
                    bgp_speaker_id=bgp_sp_id,
                    router_id=r_id)
                context.session.add(bgp_router_db)
        except db_exc.DBDuplicateEntry:
            raise l3_exc.RouterInUse(
                router_id=r_id,
                reason='is already associated with bgp speaker')
        except db_exc.DBReferenceError:
            raise l3_exc.RouterNotFound(router_id=r_id)

    @log_helpers.log_method_call
    def get_router_associated_with_bgp_speaker(self, context, bgp_sp_id):
        """Gets router associated with a bgp speaker."""
        r_id = None
        try:
            query = self._model_query(context,
                                      model.BgpSpeakerRouterAssociation)
            bsra = query.filter(
                model.BgpSpeakerRouterAssociation.bgp_speaker_id ==
                bgp_sp_id).one()
            r_id = bsra['router_id']

        except exc.NoResultFound:
            LOG.debug("the bgp speaker %s is not attached to any router",
                      bgp_sp_id)
        return r_id

    @log_helpers.log_method_call
    def get_bgp_speaker_associated_with_router(self, context, router_id):
        """Gets router associated with a bgp speaker."""
        bgp_sp_id = None
        try:
            query = self._model_query(context,
                                      model.BgpSpeakerRouterAssociation)
            bsra = query.filter(
                model.BgpSpeakerRouterAssociation.router_id == router_id).one()
            bgp_sp_id = bsra['bgp_speaker_id']

        except exc.NoResultFound:
            LOG.debug("the router %s is not attached to any bgp speaker",
                      bgp_sp_id)
        return bgp_sp_id

    @log_helpers.log_method_call
    def set_router_for_bgp_speaker_by_network(self, context,
                                              bgp_sp_id, net_id):
        """This method selects one router to be a bgp speaker.

        To pare down routers, only first subnet in specified external
        network is used for selection.
        REVISIT(Kengo): There may be a case where there are two subnets
        and a router is associated with the second subnet. The case is
        a restriction for user and should be improved later.
        """
        core_plugin = directory.get_plugin()
        subnets = core_plugin.get_subnets_by_network(context, net_id)
        if not subnets:
            raise bsri.NoSubnetInNetwork(network_id=net_id)
        if not subnets[0]['gateway_ip']:
            raise bsri.NoGatewayIpOnSubnet(subnet_id=subnets[0]['id'])
        filters = {'fixed_ips': {'ip_address': [subnets[0]['gateway_ip']],
                                 'subnet_id': [subnets[0]['id']]}}
        ports = core_plugin.get_ports(context, filters=filters)
        if not ports:
            raise bsri.NoGatewayIpPortOnSubnet(subnet_id=subnets[0]['id'])
        router_id = ports[0]['device_id']
        # If the router is already associated with bgp-speaker,
        # RouterInUse will be raised.
        self.set_router_for_bgp_speaker(context, bgp_sp_id, router_id)

    def _get_bgp_speakers_by_bgp_peer_binding(self, context, bgp_peer_id):
        with db_api.context_manager.reader.using(context):
            query = context.session.query(bdb.BgpSpeaker)
            query = query.filter(
                bdb.BgpSpeakerPeerBinding.bgp_speaker_id == bdb.BgpSpeaker.id,
                bdb.BgpSpeakerPeerBinding.bgp_peer_id == bgp_peer_id)
            return query.all()

    def delete_bgp_speaker_router_insertion(self, context, bsp_id):
        with db_api.context_manager.writer.using(context):
            query = self._model_query(
                context, model.BgpSpeakerRouterAssociation)
            query.filter(
                model.BgpSpeakerRouterAssociation.bgp_speaker_id ==
                bsp_id).delete()

    @registry.receives(resources.ROUTER, [events.BEFORE_DELETE])
    def bgp_speaker_callback(self, resource, event, trigger, **kwargs):
        router_id = kwargs['router_id']
        context = kwargs.get('context')
        if self.get_bgp_speaker_associated_with_router(context, router_id):
            raise l3_exc.RouterInUse(
                router_id=router_id, reason='is associated with bgp speaker')



# Copyright 2015 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add port binding

Revision ID: 3fe2bca71c72
Revises: 19808c5df22a
Create Date: 2015-02-16 05:24:01.270141

"""

# revision identifiers, used by Alembic.
revision = '3fe2bca71c72'
down_revision = '19808c5df22a'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'midonet_port_bindings',
        sa.Column('port_id', sa.String(length=36), sa.ForeignKey('ports.id'),
                  nullable=False, primary_key=True),
        sa.Column('interface_name', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['port_id'], ['portbindingports.port_id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('port_id'))



# Derived from:
#   nova/network/linux_net.py
#   vif_plug_ovs/linux_net.py
#
# Copyright (c) 2011 X.commerce, a business unit of eBay Inc.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import excutils

from midonet.os_vif import privsep


LOG = logging.getLogger(__name__)


def execute(*cmd, **kwargs):
    return processutils.execute(*cmd, **kwargs)


def device_exists(device):
    """Check if ethernet device exists."""
    return os.path.exists('/sys/class/net/%s' % device)


@privsep.vif_plug.entrypoint
def create_tap_dev(dev, mac_address=None, multiqueue=False):
    if not device_exists(dev):
        try:
            # First, try with 'ip'
            cmd = ('ip', 'tuntap', 'add', dev, 'mode', 'tap')
            if multiqueue:
                cmd = cmd + ('multi_queue', )
            execute(*cmd, check_exit_code=[0, 2, 254])
        except processutils.ProcessExecutionError:
            if multiqueue:
                LOG.warning(
                    'Failed to create a tap device with ip tuntap. '
                    'tunctl does not support creation of multi-queue '
                    'enabled devices, skipping fallback.')
                raise

            # Second option: tunctl
            execute('tunctl', '-b', '-t', dev)
        if mac_address:
            execute('ip', 'link', 'set', dev, 'address', mac_address,
                    check_exit_code=[0, 2, 254])
        execute('ip', 'link', 'set', dev, 'up', check_exit_code=[0, 2, 254])


@privsep.vif_plug.entrypoint
def delete_net_dev(dev):
    """Delete a network device only if it exists."""
    if device_exists(dev):
        try:
            execute('ip', 'link', 'delete', dev, check_exit_code=[0, 2, 254])
            LOG.debug("Net device removed: '%s'", dev)
        except processutils.ProcessExecutionError:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed removing net device: '%s'", dev)



# Copyright 2015 Midokura SARL
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import port as port_def
from neutron_lib.api.definitions import portbindings
from neutron_lib.api import validators
from neutron_lib.db import model_base
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory
import sqlalchemy as sa
from sqlalchemy import orm

from neutron.db import _resource_extend as resource_extend
from neutron.db import models_v2


class PortBindingInfo(model_base.BASEV2):
    __tablename__ = 'midonet_port_bindings'

    port_id = sa.Column(sa.String(36),
                        sa.ForeignKey('ports.id'),
                        sa.ForeignKey('portbindingports.port_id',
                                      ondelete="CASCADE"),
                        primary_key=True)
    interface_name = sa.Column(sa.String(length=255), nullable=False)
    port = orm.relationship(models_v2.Port,
                            backref=orm.backref("port_binding_info",
                                                lazy='joined',
                                                uselist=False,
                                                cascade='delete'))


@resource_extend.has_resource_extenders
class MidonetPortBindingMixin(object):

    def _extend_mido_portbinding(self, port_res, if_name):
        if if_name:
            port_res[portbindings.PROFILE] = {"interface_name": if_name}
        else:
            port_res[portbindings.PROFILE] = {}

    def _process_mido_portbindings_create_and_update(self, context, port_data,
                                                     port):

        port_id = port['id']

        # Set profile to {} if the binding:profile key exists but set to None.
        # This is for the special handling in the case the user wants to remove
        # the binding.
        profile = None
        if portbindings.PROFILE in port_data:
            profile = port_data.get(portbindings.PROFILE) or {}
        profile_set = validators.is_attr_set(profile)

        if_name = profile.get('interface_name') if profile_set else None
        if profile_set and profile:
            # Update or create, so validate the inputs
            if not if_name:
                msg = 'The interface name was not provided or empty'
                raise n_exc.BadRequest(resource='port', msg=msg)

            if self.get_port_host(context, port_id) is None:
                msg = 'Cannot set binding because the host is not bound'
                raise n_exc.BadRequest(resource='port', msg=msg)

        with context.session.begin(subtransactions=True):
            bind_port = context.session.query(PortBindingInfo).filter_by(
                port_id=port_id).first()
            if profile_set:
                if bind_port:
                    if if_name:
                        bind_port.interface_name = if_name
                    else:
                        context.session.delete(bind_port)
                elif if_name:
                    context.session.add(PortBindingInfo(
                        port_id=port_id, interface_name=if_name))
            else:
                if_name = bind_port.interface_name if bind_port else None

        self._extend_mido_portbinding(port, if_name)

    @staticmethod
    @resource_extend.extends([port_def.COLLECTION_NAME])
    def _extend_port_mido_portbinding(port_res, port_db):
        bind_port = port_db.port_binding_info
        if_name = bind_port.interface_name if bind_port else None
        plugin = directory.get_plugin()
        plugin._extend_mido_portbinding(port_res, if_name)



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from sqlalchemy.orm import sessionmaker

from neutron.db import api as db_api
from neutron.tests.unit import testlib_api

from midonet.neutron.db import data_state_db
from midonet.neutron.db import data_version_db as dv_db


class TestMidonetDataState(testlib_api.SqlTestCase):

    def setUp(self):
        super(TestMidonetDataState, self).setUp()
        self.session = self.get_session()
        self.session.add(data_state_db.DataState(
            updated_at=datetime.datetime.utcnow(),
            readonly=False))

    def get_session(self):
        engine = db_api.context_manager.get_legacy_facade().get_engine()
        Session = sessionmaker(bind=engine)
        return Session()

    def test_data_show(self):
        ds = data_state_db.get_data_state(self.session)
        self.assertIsNotNone(ds.id)

    def test_data_state_readonly(self):
        data_state_db.set_readonly(self.session)
        ds = data_state_db.get_data_state(self.session)
        self.assertTrue(ds.readonly)
        # TODO(Joe) - creating tasks should fail here. Implement
        # with further task_db changes coming in data sync
        data_state_db.set_readwrite(self.session)
        ds = data_state_db.get_data_state(self.session)
        self.assertFalse(ds.readonly)


class TestMidonetDataVersion(testlib_api.SqlTestCase):

    def get_session(self):
        engine = db_api.context_manager.get_legacy_facade().get_engine()
        Session = sessionmaker(bind=engine)
        return Session()

    def test_create_version(self):
        session = self.get_session()
        dv_db.create_data_version(session)
        version = dv_db.get_last_version(session)
        self.assertEqual(1, version.id)
        self.assertEqual(dv_db.STARTED, version.sync_tasks_status)

    def _test_version_status(self, version_update_func, status):
        session = self.get_session()
        dv_db.create_data_version(session)
        version = dv_db.get_last_version(session)
        self.assertEqual(dv_db.STARTED, version.sync_tasks_status)
        version_update_func(session)
        version = dv_db.get_last_version(session)
        self.assertEqual(status, version.sync_tasks_status)

    def test_update_version_status_completed(self):
        self._test_version_status(dv_db.complete_last_version,
                                  dv_db.COMPLETED)

    def test_update_version_status_error(self):
        self._test_version_status(dv_db.error_last_version,
                                  dv_db.ERROR)

    def test_update_version_status_aborted(self):
        self._test_version_status(dv_db.abort_last_version,
                                  dv_db.ABORTED)



# Copyright (C) 2016 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import excutils

from neutron_lib.plugins import constants as const
from neutron_lib.plugins import directory

from midonet.neutron.client import base as c_base
from midonet.neutron.common import config  # noqa
from midonet.neutron.common import constants as midonet_const
from midonet.neutron.db import logging_resource_db as log_res_db

LOG = logging.getLogger(__name__)


class MidonetLoggingResourcePlugin(log_res_db.LoggingResourceDbMixin):

    """Implements Logging Resource service plugin for Midonet."""

    supported_extension_aliases = ["logging-resource"]

    def __init__(self):
        super(MidonetLoggingResourcePlugin, self).__init__()

        # Instantiate MidoNet API client
        self.client = c_base.load_client(cfg.CONF.MIDONET)

    def get_plugin_type(self):
        return midonet_const.LOGGING_RESOURCE

    def get_plugin_description(self):
        """Returns string description of the plugin."""
        return "Midonet Logging Resource Service Plugin"

    @log_helpers.log_method_call
    def update_logging_resource(self, context, id, logging_resource):
        backup = self.get_logging_resource(
            context, id, fields=['name', 'description', 'enabled'])
        backup_body = {'logging_resource': backup}
        with context.session.begin(subtransactions=True):
            has_logs = self._logging_resource_has_logs(context, id)
            log_res = super(
                MidonetLoggingResourcePlugin,
                self).update_logging_resource(context, id, logging_resource)
            if has_logs:
                self.client.update_logging_resource_precommit(
                    context, id, log_res)

        if not has_logs:
            return log_res
        try:
            self.client.update_logging_resource_postcommit(id, log_res)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to update a logging "
                          "resource %(log_res_id)s in MidoNet: %(err)s",
                          {"log_res_id": log_res["id"], "err": ex})
                try:
                    super(
                        MidonetLoggingResourcePlugin,
                        self).update_logging_resource(
                            context, log_res['id'], backup_body)
                except Exception:
                    LOG.exception("Failed to update a logging resource "
                                  "for rollback %s", log_res["id"])
        return log_res

    @log_helpers.log_method_call
    def delete_logging_resource(self, context, id):
        with context.session.begin(subtransactions=True):
            super(MidonetLoggingResourcePlugin,
                  self).delete_logging_resource(context, id)
            self.client.delete_logging_resource_precommit(context, id)

        self.client.delete_logging_resource_postcommit(id)

    @log_helpers.log_method_call
    def create_logging_resource_firewall_log(self, context,
                                             firewall_log,
                                             logging_resource_id):
        with context.session.begin(subtransactions=True):
            f_log = super(
                MidonetLoggingResourcePlugin,
                self).create_logging_resource_firewall_log(
                    context, firewall_log, logging_resource_id)
            f_log_info = self._make_info_for_midonet(
                context, f_log, logging_resource_id)
            self.client.create_firewall_log_precommit(context, f_log_info)

        try:
            self.client.create_firewall_log_postcommit(f_log_info)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create a firewall log %(f_log_id)s "
                          "for %(log_res_id)s in Midonet:%(err)s",
                          {"f_log_id": f_log["id"],
                           "log_res_id": logging_resource_id, "err": ex})
                try:
                    super(
                        MidonetLoggingResourcePlugin,
                        self).delete_logging_resource_firewall_log(
                            context, f_log["id"], logging_resource_id)
                except Exception:
                    LOG.exception("Failed to delete a firewall_log %s",
                                  f_log["id"])

        return f_log

    @log_helpers.log_method_call
    def update_logging_resource_firewall_log(
            self, context, id, logging_resource_id, firewall_log):
        backup = self.get_logging_resource_firewall_log(
            context, id, logging_resource_id,
            fields=['fw_event', 'description'])
        backup_body = {'firewall_log': backup}
        with context.session.begin(subtransactions=True):
            f_log = super(
                MidonetLoggingResourcePlugin,
                self).update_logging_resource_firewall_log(
                    context, id, logging_resource_id, firewall_log)
            f_log_info = self._make_info_for_midonet(
                context, f_log, logging_resource_id)
            self.client.update_firewall_log_precommit(context, id, f_log_info)

        try:
            self.client.update_firewall_log_postcommit(id, f_log_info)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to update a firewall log "
                          "%(f_log_id)s in Midonet:%(err)s",
                          {"f_log_id": f_log["id"], "err": ex})
                try:
                    super(
                        MidonetLoggingResourcePlugin,
                        self).update_logging_resource_firewall_log(
                            context, f_log['id'],
                            logging_resource_id, backup_body)
                except Exception:
                    LOG.exception("Failed to update a firewall log "
                                  "for rollback %s", f_log["id"])
        return f_log

    @log_helpers.log_method_call
    def delete_logging_resource_firewall_log(self, context, id,
                                             logging_resource_id):
        with context.session.begin(subtransactions=True):
            super(MidonetLoggingResourcePlugin,
                  self).delete_logging_resource_firewall_log(
                context, id, logging_resource_id)
            self.client.delete_firewall_log_precommit(context, id)

        self.client.delete_firewall_log_postcommit(id)

    def _make_info_for_midonet(self, context, f_log, logging_resource_id):
        f_log_info = f_log.copy()
        # Get firewall and logging resource object for backend.
        fw_plugin = directory.get_plugin(const.FIREWALL)
        f_log_info['firewall'] = fw_plugin.get_firewall(
            context, f_log_info['firewall_id'], fields=['id', 'tenant_id'])
        f_log_info['logging_resource'] = self.get_logging_resource(
            context, logging_resource_id)
        del f_log_info['logging_resource']['firewall_logs']

        return f_log_info



# Copyright 2014: Intel Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Modifications Copyright 2016: Midokura

from rally import consts
from rally.plugins.openstack import scenario
from rally.plugins.openstack.scenarios.neutron import utils
from rally.task import atomic
from rally.task import validation


@validation.add("required_services", services=[consts.Service.NEUTRON])
@validation.add("required_platform", platform="openstack", users=True)
@scenario.configure(context={"cleanup@openstack": ["neutron"]},
                    name="NetworkPlugin.create_networks")
class CreateNetwork(utils.NeutronScenario):
    """Benchmark scenarios for Neutron."""

    def run(self, network_create_args=None):
        """Create a network.

        Measure the "neutron net-create" command performance.

        If you have only 1 user in your context, you will
        add 1 network on every iteration.

        :param network_create_args: dict, POST /v2.0/networks request options
        """
        self._create_network(network_create_args or {})


@validation.add("number", param_name="subnets_per_network",
                minval=1, integer_only=True)
@validation.add("required_services", services=[consts.Service.NEUTRON])
@validation.add("required_platform", platform="openstack", users=True)
@scenario.configure(context={"cleanup@openstack": ["neutron"]},
                    name="NetworkPlugin.create_subnets")
class CreateSubnets(utils.NeutronScenario):
    def run(self,
            network_create_args=None,
            subnet_create_args=None,
            subnet_cidr_start=None,
            subnets_per_network=None):
        """Create a given number of subnets.

        The scenario creates a network and a given number of subnets.

        :param network_create_args: dict, POST /v2.0/networks request
                                    options. Deprecated
        :param subnet_create_args: dict, POST /v2.0/subnets request options
        :param subnet_cidr_start: str, start value for subnets CIDR
        :param subnets_per_network: int, number of subnets for one network
        """
        network = self._get_or_create_network(network_create_args)
        self._create_subnets(network, subnet_create_args, subnet_cidr_start,
                             subnets_per_network)


@validation.add("number", param_name="subnets_per_network",
                minval=1, integer_only=True)
@validation.add("required_services", services=[consts.Service.NEUTRON])
@validation.add("required_platform", platform="openstack", users=True)
@scenario.configure(context={"cleanup@openstack": ["neutron"]},
                    name="NetworkPlugin.create_subnets_routers_interfaces")
class CreateSubnetsRoutersInterfaces(utils.NeutronScenario):
    def run(self,
            network_create_args=None,
            subnet_create_args=None,
            subnet_cidr_start=None,
            subnets_per_network=None,
            router_create_args=None):
        """
        Create a network, a given number of subnets and routers.

        :param network_create_args: dict, POST /v2.0/networks request
                                    options. Deprecated.
        :param subnet_create_args: dict, POST /v2.0/subnets request options
        :param subnet_cidr_start: str, start value for subnets CIDR
        :param subnets_per_network: int, number of subnets for one network
        :param router_create_args: dict, POST /v2.0/routers request options
        """
        self._create_network_structure(network_create_args, subnet_create_args,
                                       subnet_cidr_start, subnets_per_network,
                                       router_create_args)


@validation.add("number", param_name="routers_per_subnet",
                minval=1, integer_only=True)
@validation.add("required_services", services=[consts.Service.NEUTRON])
@validation.add("required_platform", platform="openstack", users=True)
@scenario.configure(context={"cleanup@openstack": ["neutron"]},
                    name="NetworkPlugin.create_routers")
class CreateRouters(utils.NeutronScenario):
    def run(self,
            network_create_args=None,
            routers_per_subnet=None,
            router_create_args=None,
            port_create_args=None):
        """
        Create a given number of ports and routers.

        :param network_create_args: dict, POST /v2.0/networks request
                                    options. Deprecated.
        :param routers_per_subnet: int, number of routers for one subnet
        :param router_create_args: dict, POST /v2.0/routers request options
        :param port_create_args: dict, POST /v2.0/ports request options
        """

        network = self._get_or_create_network(network_create_args)
        for i in range(routers_per_subnet):
            router = self._create_router(router_create_args or {})
            port = self._create_port(network, port_create_args or {})
            self._add_interface_router_port(router, port)

    @atomic.action_timer("neutron.add_interface_router_port")
    def _add_interface_router_port(self, router, port):
        self.clients("neutron").add_interface_router(
            router["router"]["id"], {"port_id": port["port"]["id"]})


@validation.add("number", param_name="ports_per_network",
                minval=1, integer_only=True)
@validation.add("required_services", services=[consts.Service.NEUTRON])
@validation.add("required_platform", platform="openstack", users=True)
@scenario.configure(context={"cleanup@openstack": ["neutron"]},
                    name="NetworkPlugin.create_ports")
class CreatePorts(utils.NeutronScenario):
    def run(self,
            network_create_args=None,
            port_create_args=None,
            ports_per_network=None):
        """Create a given number of ports.

        :param network_create_args: dict, POST /v2.0/networks request
                                    options. Deprecated.
        :param port_create_args: dict, POST /v2.0/ports request options
        :param ports_per_network: int, number of ports for one network
        """
        network = self._get_or_create_network(network_create_args)
        for i in range(ports_per_network):
            self._create_port(network, port_create_args or {})



# Copyright (C) 2016 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

from oslo_utils import uuidutils
import webob.exc

from neutron.tests.unit.api import test_extensions as test_ex
from neutron.tests.unit.extensions import test_l3
from neutron_fwaas.tests.unit.services.firewall import test_fwaas_plugin as tfp

from midonet.neutron import extensions as midoextensions
from midonet.neutron.extensions import logging_resource as log_res_ext
from midonet.neutron.tests.unit import test_midonet_plugin_ml2 as test_mn_ml2


LOGGING_PLUGIN_KLASS = 'midonet_logging_resource'
extensions_path = ':'.join(midoextensions.__path__)

FAKE_LOG_RES_NAME = 'log_res'
NEW_LOG_RES_NAME = 'new_log_res'
FAKE_LOG_RES_DESC = 'log_res_description'
NEW_LOG_RES_DESC = 'new_log_res_description'
ENABLED_TRUE = True
ENABLED_FALSE = False
FAKE_FW_LOG_DESC = 'fw_log_description'
NEW_FW_LOG_DESC = 'new_fw_log_description'
NOT_FOUND_FW_UUID = 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
NOT_FOUND_LOG_RES_UUID = 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
NOT_FOUND_FW_LOG_UUID = 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'


class LoggingResourceTestExtensionManager(tfp.FirewallTestExtensionManager):

    def get_resources(self):
        res = super(LoggingResourceTestExtensionManager,
                    self).get_resources()
        return res + log_res_ext.Logging_resource.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class LoggingResourceTestCaseMixin(test_l3.L3NatTestCaseMixin):

    def setUp(self, plugin=None, service_plugins=None, ext_mgr=None):

        service_plugins = {
            'logging_resource_plugin_name': LOGGING_PLUGIN_KLASS,
            'fwaas_plugin_name': 'midonet_firewall'}

        log_res_mgr = LoggingResourceTestExtensionManager()
        super(LoggingResourceTestCaseMixin, self).setUp(
            service_plugins=service_plugins, ext_mgr=log_res_mgr)
        self.ext_api = test_ex.setup_extensions_middleware(log_res_mgr)

        router1 = self._make_router(self.fmt, uuidutils.generate_uuid(),
                                    'router1', True)
        self._router_id1 = router1['router']['id']
        router2 = self._make_router(self.fmt, uuidutils.generate_uuid(),
                                    'router2', True)
        self._router_id2 = router2['router']['id']
        self._tenant_id1 = uuidutils.generate_uuid()
        self._tenant_id2 = uuidutils.generate_uuid()
        fw1 = self._create_firewall(self._tenant_id1)
        self._fw_id1 = fw1['firewall']['id']
        fw2 = self._create_firewall(self._tenant_id1)
        self._fw_id2 = fw2['firewall']['id']

    @contextlib.contextmanager
    def firewall(self, tenant_id):
        firewall = self._create_firewall(tenant_id)
        yield firewall

    def _create_firewall(self, tenant_id):
        fw_policy_data = {'firewall_policy': {'tenant_id': tenant_id,
                                              'name': 'fw_policy'}}
        req = self.new_create_request('fw/firewall_policies', fw_policy_data,
                                      'json')
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        firewall_data = {'firewall': {
            'name': 'fw',
            'firewall_policy_id': res['firewall_policy']['id'],
            'tenant_id': tenant_id,
            'shared': False}}
        req = self.new_create_request('fw/firewalls', firewall_data, 'json')
        return self.deserialize(self.fmt, req.get_response(self.ext_api))

    @contextlib.contextmanager
    def logging_resource(self, name=FAKE_LOG_RES_NAME, tenant_id=None,
                         desc=FAKE_LOG_RES_DESC, enabled=ENABLED_FALSE):
        logging_resource = self._make_logging_resource(name, tenant_id,
                                                       desc, enabled)
        yield logging_resource

    def _make_logging_resource(self, name, tenant_id, desc, enabled):
        res = self._create_logging_resource(name, tenant_id, desc, enabled)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    def _create_logging_resource(self, name=FAKE_LOG_RES_NAME, tenant_id=None,
                                 desc=FAKE_LOG_RES_DESC,
                                 enabled=ENABLED_FALSE):
        data = {'logging_resource': {'tenant_id': tenant_id or self._tenant_id,
                                     'name': name,
                                     'description': desc,
                                     'enabled': enabled}}
        log_res_req = self.new_create_request('logging/logging_resources',
                                              data, self.fmt)
        return log_res_req.get_response(self.ext_api)

    @contextlib.contextmanager
    def firewall_log(self, log_res_id, firewall_id=None,
                     desc=FAKE_FW_LOG_DESC, tenant_id=None,
                     fw_event=log_res_ext.FW_EVENT_ALL):
        firewall_log = self._make_firewall_log(log_res_id, firewall_id,
                                               desc, tenant_id, fw_event)
        yield firewall_log

    def _make_firewall_log(self, log_res_id, firewall_id, desc,
                           tenant_id, fw_event):
        res = self._create_firewall_log(log_res_id, firewall_id,
                                        desc, tenant_id, fw_event)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    def _create_firewall_log(self, log_res_id, firewall_id=None,
                             desc=FAKE_FW_LOG_DESC, tenant_id=None,
                             fw_event=log_res_ext.FW_EVENT_ALL):
        data = {'firewall_log': {'tenant_id': tenant_id or self._tenant_id,
                                 'description': desc,
                                 'fw_event': fw_event}}
        if firewall_id:
            data['firewall_log']['firewall_id'] = firewall_id
        f_log_req = self.new_create_request('logging/logging_resources/'
                                            + log_res_id + '/firewall_logs',
                                            data, self.fmt)
        return f_log_req.get_response(self.ext_api)

    def test_create_logging_resource(self):
        expected = {'name': FAKE_LOG_RES_NAME,
                    'description': FAKE_LOG_RES_DESC,
                    'enabled': ENABLED_FALSE,
                    'firewall_logs': []}
        with self.logging_resource() as log_res:
            self.assertDictSupersetOf(expected, log_res['logging_resource'])

    def test_show_logging_resource(self):
        expected = {'name': FAKE_LOG_RES_NAME,
                    'description': FAKE_LOG_RES_DESC,
                    'enabled': ENABLED_FALSE,
                    'firewall_logs': []}
        with self.logging_resource() as log_res:
            req = self.new_show_request('logging/logging_resources',
                                        log_res['logging_resource']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['logging_resource'])

    def test_show_logging_resource_not_found(self):
        req = self.new_show_request('logging/logging_resources',
                                    NOT_FOUND_LOG_RES_UUID)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_list_logging_resource(self):
        with self.logging_resource(), self.logging_resource():
            req = self.new_list_request('logging/logging_resources')
            res = self.deserialize(
                self.fmt, req.get_response(self.ext_api))
            self.assertEqual(2, len(res['logging_resources']))

    def test_update_logging_resource_without_firewall_log(self):
        expected = {'name': NEW_LOG_RES_NAME,
                    'description': NEW_LOG_RES_DESC,
                    'enabled': ENABLED_TRUE}
        with self.logging_resource() as log_res:
            data = {'logging_resource': {'name': NEW_LOG_RES_NAME,
                                         'description': NEW_LOG_RES_DESC,
                                         'enabled': ENABLED_TRUE}}
            req = self.new_update_request('logging/logging_resources',
                                          data,
                                          log_res['logging_resource']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['logging_resource'])
            self.client_mock.update_logging_resource_postcommit. \
                assert_not_called()

    def test_update_logging_resource_with_firewall_log(self):
        expected = {'name': NEW_LOG_RES_NAME,
                    'description': NEW_LOG_RES_DESC,
                    'enabled': ENABLED_TRUE}
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1):
            data = {'logging_resource': {'name': NEW_LOG_RES_NAME,
                                         'description': NEW_LOG_RES_DESC,
                                         'enabled': ENABLED_TRUE}}
            req = self.new_update_request('logging/logging_resources',
                                          data,
                                          log_res['logging_resource']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['logging_resource'])
            self.client_mock.update_logging_resource_postcommit.assert_called()

    def test_update_logging_resource_error_rollback_neutron_resource(self):
        self.client_mock.update_logging_resource_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1):
            data = {'logging_resource': {'enabled': ENABLED_TRUE}}
            req = self.new_update_request('logging/logging_resources',
                                          data,
                                          log_res['logging_resource']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            # ensure enabled are not changed.
            expected = {'enabled': ENABLED_FALSE}
            req = self.new_show_request('logging/logging_resources',
                                        log_res['logging_resource']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['logging_resource'])

    def test_delete_logging_resource(self):
        with self.logging_resource() as log_res:
            req = self.new_delete_request('logging/logging_resources',
                                          log_res['logging_resource']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_delete_logging_resource_with_firewall_logs(self):
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1), \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id2):
            req = self.new_delete_request('logging/logging_resources',
                                          log_res['logging_resource']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_delete_logging_resource_error_delete_neutron_resource(self):
        self.client_mock.delete_logging_resource_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.logging_resource() as log_res:
            req = self.new_delete_request('logging/logging_resources',
                                          log_res['logging_resource']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            # check the resource deleted in Neutron DB
            req = self.new_list_request('logging/logging_resources')
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertFalse(res['logging_resources'])

    def test_create_firewall_log(self):
        expected = {'description': FAKE_FW_LOG_DESC,
                    'fw_event': log_res_ext.FW_EVENT_ALL,
                    'firewall_id': self._fw_id1}
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1) as f_log:
            self.assertDictSupersetOf(expected, f_log['firewall_log'])

    def test_create_firewall_log_diff_log_res_diff_tenant_same_firewall(self):
        with self.logging_resource() as log_res, \
                self.logging_resource(
                    tenant_id=self._tenant_id2) as log_res2, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1):
            res = self._create_firewall_log(
                log_res2['logging_resource']['id'],
                firewall_id=self._fw_id1,
                tenant_id=self._tenant_id2)
            self.assertEqual(webob.exc.HTTPCreated.code, res.status_int)

    def test_create_firewall_log_with_not_found_firewall(self):
        with self.logging_resource() as log_res:
            res = self._create_firewall_log(
                log_res['logging_resource']['id'],
                firewall_id=NOT_FOUND_FW_UUID)
            self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_create_firewall_log_in_same_log_res_with_same_firewall(self):
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1):
            res = self._create_firewall_log(
                log_res['logging_resource']['id'],
                firewall_id=self._fw_id1)
            self.assertEqual(webob.exc.HTTPCreated.code, res.status_int)

    def test_create_firewall_log_in_diff_log_res_with_same_firewall(self):
        with self.logging_resource() as log_res, \
                self.logging_resource() as log_res2, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1):
            res = self._create_firewall_log(
                log_res2['logging_resource']['id'],
                firewall_id=self._fw_id1)
            self.assertEqual(webob.exc.HTTPCreated.code, res.status_int)

    def test_create_firewall_log_error_delete_neutron_resource(self):
        self.client_mock.create_firewall_log_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.logging_resource() as log_res:
            try:
                with self.firewall_log(log_res['logging_resource']['id']):
                    self.assertTrue(False)
            except webob.exc.HTTPClientError:
                pass
            req = self.new_list_request('logging/logging_resources/'
                                        + log_res['logging_resource']['id']
                                        + '/firewall_logs')
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertFalse(res['firewall_logs'])

    def test_update_firewall_log(self):
        expected = {'description': NEW_FW_LOG_DESC,
                    'fw_event': log_res_ext.FW_EVENT_DROP}
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1) as f_log:
            data = {'firewall_log': {'description': NEW_FW_LOG_DESC,
                                     'fw_event': log_res_ext.FW_EVENT_DROP}}
            req = self.new_update_request('logging/logging_resources/'
                                          + log_res['logging_resource']['id']
                                          + '/firewall_logs',
                                          data,
                                          f_log['firewall_log']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['firewall_log'])

    def test_update_firewall_log_error_rollback_neutron_resource(self):
        self.client_mock.update_firewall_log_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1) as f_log:
            data = {'firewall_log': {'fw_event': log_res_ext.FW_EVENT_DROP}}
            req = self.new_update_request('logging/logging_resources/'
                                          + log_res['logging_resource']['id']
                                          + '/firewall_logs',
                                          data,
                                          f_log['firewall_log']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            # ensure fw_event is not changed.
            expected = {'fw_event': log_res_ext.FW_EVENT_ALL}
            req = self.new_show_request('logging/logging_resources/'
                                        + log_res['logging_resource']['id']
                                        + '/firewall_logs',
                                        f_log['firewall_log']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['firewall_log'])

    def test_show_firewall_log(self):
        expected = {'description': FAKE_FW_LOG_DESC,
                    'fw_event': log_res_ext.FW_EVENT_ALL,
                    'firewall_id': self._fw_id1}
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1) as f_log:
            req = self.new_show_request('logging/logging_resources/'
                                        + log_res['logging_resource']['id']
                                        + '/firewall_logs',
                                        f_log['firewall_log']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['firewall_log'])

    def test_show_firewall_log_not_found(self):
        with self.logging_resource() as log_res:
            req = self.new_show_request('logging/logging_resources/'
                                        + log_res['logging_resource']['id']
                                        + '/firewall_logs',
                                        NOT_FOUND_FW_LOG_UUID)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_list_firewall_log(self):
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1), \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id2):
            req = self.new_list_request('logging/logging_resources/'
                                        + log_res['logging_resource']['id']
                                        + '/firewall_logs')
            res = self.deserialize(
                self.fmt, req.get_response(self.ext_api))
            self.assertEqual(2, len(res['firewall_logs']))

    def test_delete_firewall_log(self):
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1) as f_log:
            req = self.new_delete_request('logging/logging_resources/'
                                          + log_res['logging_resource']['id']
                                          + '/firewall_logs',
                                          f_log['firewall_log']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_delete_firewall_log_error_delete_neutron_resource(self):
        self.client_mock.delete_firewall_log_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.logging_resource() as log_res, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=self._fw_id1) as f_log:
            req = self.new_delete_request('logging/logging_resources/'
                                          + log_res['logging_resource']['id']
                                          + '/firewall_logs',
                                          f_log['firewall_log']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            # check the resource deleted in Neutron DB
            req = self.new_list_request('logging/logging_resources')
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(
                [],
                res['logging_resources'][0]['firewall_logs'])

    def test_delete_firewall_with_firewall_log(self):
        with self.logging_resource() as log_res, \
                self.firewall(self._tenant_id) as fw, \
                self.firewall_log(log_res['logging_resource']['id'],
                                  firewall_id=fw['firewall']['id']) as f_log:
            req = self.new_delete_request('fw/firewalls',
                                          fw['firewall']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
            req = self.new_show_request('logging/logging_resources/'
                                        + log_res['logging_resource']['id']
                                        + '/firewall_logs',
                                        f_log['firewall_log']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_delete_firewall_with_multiple_firewall_logs(self):
        with self.logging_resource() as log_res1, \
                self.logging_resource(
                    tenant_id=self._tenant_id2) as log_res2, \
                self.firewall(self._tenant_id) as fw, \
                self.firewall_log(
                    log_res1['logging_resource']['id'],
                    firewall_id=fw['firewall']['id']) as f_log1, \
                self.firewall_log(log_res2['logging_resource']['id'],
                                  firewall_id=fw['firewall']['id'],
                                  tenant_id=self._tenant_id2) as f_log2:
            req = self.new_delete_request('fw/firewalls',
                                          fw['firewall']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
            req = self.new_show_request('logging/logging_resources/'
                                        + log_res1['logging_resource']['id']
                                        + '/firewall_logs',
                                        f_log1['firewall_log']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)
            req = self.new_show_request('logging/logging_resources/'
                                        + log_res2['logging_resource']['id']
                                        + '/firewall_logs',
                                        f_log2['firewall_log']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)


class LoggingResourceTestCaseML2(LoggingResourceTestCaseMixin,
                                 test_mn_ml2.MidonetPluginML2TestCase):
    pass



# Copyright (C) 2013 Midokura PTE LTD
# Copyright (C) 2014 Midokura SARL.
# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.agent.linux import dhcp


class DhcpNoOpDriver(dhcp.DhcpLocalProcess):

    @classmethod
    def existing_dhcp_networks(cls, conf):
        """Return a list of existing networks ids that we have configs for."""
        return []

    @classmethod
    def check_version(cls):
        """Execute version checks on DHCP server."""
        return float(1.0)

    def disable(self, retain_port=False):
        """Disable DHCP for this network."""
        if not retain_port:
            self.device_manager.destroy(self.network, self.interface_name)
        self._remove_config_files()

    def reload_allocations(self):
        """Force the DHCP server to reload the assignment database."""
        pass

    def spawn_process(self):
        pass

    @classmethod
    def should_enable_metadata(cls, conf, network):
        """We need MD namespace proxy regardless of network configuration"""
        return True



# Copyright (C) 2016 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

from oslo_utils import uuidutils
import webob.exc

from neutron.tests.unit.api import test_extensions as test_ex
from neutron.tests.unit.extensions import test_extraroute
from neutron.tests.unit.extensions import test_l3
from neutron_dynamic_routing.extensions import bgp

from midonet.neutron.common import constants as m_const
from midonet.neutron import extensions as midoextensions
from midonet.neutron.extensions import bgp_speaker_router_insertion as bsri
from midonet.neutron.services.bgp import plugin as bgp_plugin
from midonet.neutron.tests.unit import test_midonet_plugin_ml2 as test_mn_ml2

FAKE_SPEAKER_NAME = "bgp_spaeker_1"
FAKE_LOCAL_AS = 65000
FAKE_REMOTE_AS = 65001
NOT_FOUND_ROUTER_UUID = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
FAKE_PEER_IP = "192.168.0.254"
FAKE_PEER_IP2 = "192.168.1.254"
FAKE_PEER_NAME = "bgp_peer_1"
FAKE_PEER_NAME2 = "bgp_peer_2"
AUTH_TYPE_MD5 = "md5"
AUTH_TYPE_NONE = "none"
FAKE_AUTH_PASSWORD = "testtest"
ADD_BGP_PEER_ACTION = "add_bgp_peer"
REMOVE_BGP_PEER_ACTION = "remove_bgp_peer"
ADD_NETWORK_ACTION = "add_gateway_network"
REMOVE_NETWORK_ACTION = "remove_gateway_network"
ADD_INTERFACE_ACTION = "add_router_interface"
REMOVE_INTERFACE_ACTION = "remove_router_interface"

BGP_PLUGIN_KLASS = 'midonet_bgp'
extensions_path = ':'.join(midoextensions.__path__)


class BgpTestExtensionManager(
        test_extraroute.ExtraRouteTestExtensionManager):

    def get_resources(self):
        res = super(BgpTestExtensionManager, self).get_resources()
        bgp.RESOURCE_ATTRIBUTE_MAP['bgp-speakers'].update(
            bsri.EXTENDED_ATTRIBUTES_2_0['bgp-speakers'])
        return res + bgp.Bgp.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class BgpTestCaseMixin(test_l3.L3NatTestCaseMixin):

    def setUp(self, plugin=None, service_plugins=None, ext_mgr=None):

        service_plugins = {'bgp_plugin_name': BGP_PLUGIN_KLASS}
        bgp_mgr = BgpTestExtensionManager()
        super(BgpTestCaseMixin, self).setUp(
            service_plugins=service_plugins, ext_mgr=bgp_mgr)
        self.ext_api = test_ex.setup_extensions_middleware(bgp_mgr)
        self.bgp_plugin = bgp_plugin.MidonetBgpPlugin()

        network = self._make_network(self.fmt, 'net1', True)
        self._subnet = self._make_subnet(self.fmt, network, "192.168.0.1",
                                         '192.168.0.0/24')
        self._subnet_id = self._subnet['subnet']['id']
        network = self._make_network(self.fmt, 'net2', True)
        self._subnet = self._make_subnet(self.fmt, network, "192.168.1.1",
                                         '192.168.1.0/24')
        self._subnet_id2 = self._subnet['subnet']['id']
        network = self._make_network(self.fmt, 'net3', True)
        self._subnet = self._make_subnet(self.fmt, network, "192.168.2.1",
                                         '192.168.2.0/24')
        self._subnet_id3 = self._subnet['subnet']['id']
        port = self._make_port(self.fmt, network['network']['id'])
        self._port_id = port['port']['id']
        self._port_fixed_ip = port['port']['fixed_ips'][0]['ip_address']
        router1 = self._make_router(self.fmt, uuidutils.generate_uuid(),
                                    'router1', True)
        self._router_id = router1['router']['id']
        router2 = self._make_router(self.fmt, uuidutils.generate_uuid(),
                                    'router2', True)
        self._router_id2 = router2['router']['id']
        self._router_interface_action('add', self._router_id, self._subnet_id,
                                      None)
        self._router_interface_action('add', self._router_id, self._subnet_id2,
                                      None)
        self._router_interface_action('add', self._router_id, self._subnet_id3,
                                      None)
        self._router_interface_action('add', self._router_id2, None,
                                      self._port_id)

        # for non-router use case
        ext_net = self._make_network(self.fmt, 'ext_net2', True)
        self._set_net_external(ext_net['network']['id'])
        self._ext_net_id = ext_net['network']['id']
        self._ext_subnet = self._make_subnet(self.fmt, ext_net, "100.65.0.1",
                                             '100.65.0.0/24')
        self._ext_subnet_id = self._ext_subnet['subnet']['id']
        edge_router = self._make_router(self.fmt, uuidutils.generate_uuid(),
                                        'edge_router', True)
        self._edge_router_id = edge_router['router']['id']
        self._router_interface_action('add', self._edge_router_id,
                                      self._ext_subnet_id, None)

    @contextlib.contextmanager
    def bgp_speaker(self, name=FAKE_SPEAKER_NAME,
                    local_as=FAKE_LOCAL_AS, router_id=None):
        bgp_speaker = self._make_bgp_speaker(name, local_as, router_id)
        yield bgp_speaker

    @contextlib.contextmanager
    def bgp_peer(self, name=FAKE_PEER_NAME, peer_ip=FAKE_PEER_IP,
                 remote_as=FAKE_REMOTE_AS, auth_type=AUTH_TYPE_MD5):
        bgp_peer = self._make_bgp_peer(name, peer_ip,
                                       remote_as, auth_type)
        yield bgp_peer

    def _make_bgp_speaker(self, name, local_as, router_id):
        res = self._create_bgp_speaker(name, local_as, router_id)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    def _make_bgp_peer(self, name, peer_ip, remote_as, auth_type):
        res = self._create_bgp_peer(name, peer_ip, remote_as, auth_type)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    def _create_bgp_speaker(self, name=FAKE_SPEAKER_NAME,
                            local_as=FAKE_LOCAL_AS, router_id=None):
        data = {'bgp_speaker': {'tenant_id': uuidutils.generate_uuid(),
                                'name': name,
                                'local_as': local_as,
                                'ip_version': 4}}
        if router_id:
            data['bgp_speaker'][m_const.LOGICAL_ROUTER] = router_id
        bgp_sp_req = self.new_create_request('bgp-speakers',
                                             data, self.fmt)
        return bgp_sp_req.get_response(self.ext_api)

    def _create_bgp_peer(self, name=FAKE_PEER_NAME, peer_ip=FAKE_PEER_IP,
                         remote_as=FAKE_REMOTE_AS,
                         auth_type=AUTH_TYPE_MD5):
        data = {'bgp_peer': {'tenant_id': uuidutils.generate_uuid(),
                             'name': name,
                             'peer_ip': peer_ip,
                             'remote_as': remote_as,
                             'auth_type': auth_type}}
        if auth_type == AUTH_TYPE_MD5:
            data['bgp_peer']['password'] = FAKE_AUTH_PASSWORD
        bgp_peer_req = self.new_create_request('bgp-peers',
                                               data, self.fmt)
        return bgp_peer_req.get_response(self.ext_api)

    def test_create_bgp_speaker_with_router(self):
        expected = {'name': FAKE_SPEAKER_NAME,
                    'local_as': FAKE_LOCAL_AS,
                    'ip_version': 4,
                    m_const.LOGICAL_ROUTER: self._router_id}
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker:
            self.assertDictSupersetOf(expected, bgp_speaker['bgp_speaker'])

    def test_update_bgp_speaker_with_router(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker:
            data = {'bgp_speaker': {'name': 'new_name'}}
            req = self.new_update_request('bgp-speakers',
                                          data,
                                          bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertEqual(data['bgp_speaker']['name'],
                             res['bgp_speaker']['name'])

    def test_show_bgp_speaker_with_router(self):
        expected = {'name': FAKE_SPEAKER_NAME,
                    'local_as': FAKE_LOCAL_AS,
                    'ip_version': 4,
                    m_const.LOGICAL_ROUTER: self._router_id}
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker:
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['bgp_speaker'])

    def test_list_bgp_speaker_with_router(self):
        with self.bgp_speaker(router_id=self._router_id), \
                self.bgp_speaker(router_id=self._router_id2):
            req = self.new_list_request('bgp-speakers')
            res = self.deserialize(
                self.fmt, req.get_response(self.ext_api))
            self.assertEqual(2, len(res['bgp_speakers']))

    def test_delete_bgp_speaker(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker:
            req = self.new_delete_request('bgp-speakers',
                                          bgp_speaker['bgp_speaker']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_delete_bgp_speaker_with_peer(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            peers = [bgp_peer['bgp_peer']['id']]
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(peers, res['bgp_speaker']['peers'])
            req = self.new_delete_request('bgp-speakers',
                                          bgp_speaker['bgp_speaker']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_create_bgp_speaker_with_same_router(self):
        with self.bgp_speaker(router_id=self._router_id):
            res = self._create_bgp_speaker(router_id=self._router_id)
            self.deserialize(self.fmt, res)
            self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_create_bgp_speaker_with_not_found_router(self):
        res = self._create_bgp_speaker(router_id=NOT_FOUND_ROUTER_UUID)
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_create_bgp_speaker_without_router(self):
        expected = {'name': FAKE_SPEAKER_NAME,
                    'local_as': FAKE_LOCAL_AS,
                    'ip_version': 4}
        with self.bgp_speaker() as bgp_speaker:
            self.assertDictSupersetOf(expected, bgp_speaker['bgp_speaker'])

    def test_delete_bgp_speaker_error_delete_neutron_resource(self):
        self.client_mock.update_bgp_speaker_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            req = self.new_delete_request('bgp-speakers',
                                          bgp_speaker['bgp_speaker']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            # check the resource deleted in Neutron DB
            req = self.new_list_request('bgp-speakers')
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertFalse(res['bgp_speakers'])

    def test_delete_router_in_use_by_bgp_speaker(self):
        with self.bgp_speaker(router_id=self._router_id):
            req = self.new_delete_request('routers',
                                          self._router_id)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_create_bgp_peer(self):
        expected = {'name': FAKE_PEER_NAME,
                    'peer_ip': FAKE_PEER_IP,
                    'remote_as': FAKE_REMOTE_AS,
                    'auth_type': AUTH_TYPE_MD5}
        with self.bgp_peer() as bgp_peer:
            self.assertDictSupersetOf(expected, bgp_peer['bgp_peer'])

    def test_create_bgp_peer_with_auth_type_none(self):
        expected = {'name': FAKE_PEER_NAME,
                    'peer_ip': FAKE_PEER_IP,
                    'remote_as': FAKE_REMOTE_AS,
                    'auth_type': AUTH_TYPE_NONE}
        with self.bgp_peer(auth_type=AUTH_TYPE_NONE) as bgp_peer:
            self.assertDictSupersetOf(expected, bgp_peer['bgp_peer'])

    def test_update_bgp_peer(self):
        with self.bgp_peer() as bgp_peer:
            data = {'bgp_peer': {'password': 'new_password'}}
            req = self.new_update_request('bgp-peers',
                                          data,
                                          bgp_peer['bgp_peer']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPOk.code, res.status_int)

    def test_show_bgp_peer(self):
        expected = {'name': FAKE_PEER_NAME,
                    'remote_as': FAKE_REMOTE_AS,
                    'peer_ip': FAKE_PEER_IP,
                    'auth_type': AUTH_TYPE_MD5}
        with self.bgp_peer() as bgp_peer:
            req = self.new_show_request('bgp-peers',
                                        bgp_peer['bgp_peer']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['bgp_peer'])

    def test_list_bgp_peer(self):
        with self.bgp_peer(), self.bgp_peer():
            req = self.new_list_request('bgp-peers')
            res = self.deserialize(
                self.fmt, req.get_response(self.ext_api))
            self.assertEqual(2, len(res['bgp_peers']))

    def test_delete_bgp_peer(self):
        with self.bgp_peer() as bgp_peer:
            req = self.new_delete_request('bgp-peers',
                                          bgp_peer['bgp_peer']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_delete_bgp_peer_related_to_bgp_speaker(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            peers = [bgp_peer['bgp_peer']['id']]
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(peers, res['bgp_speaker']['peers'])
            req = self.new_delete_request('bgp-peers',
                                          bgp_peer['bgp_peer']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual([], res['bgp_speaker']['peers'])

    def test_add_bgp_peer_error_neutron_resource_rollback(self):
        self.client_mock.create_bgp_peer_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual([], res['bgp_speaker']['peers'])

    def test_add_bgp_peer_with_associated_bgp_speaker(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer, self.bgp_peer() as bgp_peer2:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            data = {"bgp_peer_id": bgp_peer2['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_add_bgp_peer_to_bgp_speaker_without_router(self):
        with self.bgp_speaker() as bgp_speaker, self.bgp_peer() as bgp_peer:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_add_bgp_peer_without_bgp_peer_id_parameter(self):
        with self.bgp_speaker() as bgp_speaker, self.bgp_peer():
            req = self.new_action_request('bgp-speakers', {},
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_remove_bgp_peer_from_speaker_with_router(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            peers = [bgp_peer['bgp_peer']['id']]
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(peers, res['bgp_speaker']['peers'])
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          REMOVE_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual([], res['bgp_speaker']['peers'])

    def test_delete_bgp_peer_error_delete_neutron_resource(self):
        self.client_mock.delete_bgp_peer_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            peers = [bgp_peer['bgp_peer']['id']]
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(peers, res['bgp_speaker']['peers'])
            req = self.new_delete_request('bgp-peers',
                                          bgp_peer['bgp_peer']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            # check the resource deleted in Neutron DB
            req = self.new_list_request('bgp-peers')
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertFalse(res['bgp_peers'])

    def test_remove_bgp_peer_error_delete_neutron_resource(self):
        self.client_mock.delete_bgp_peer_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          REMOVE_BGP_PEER_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual([], res['bgp_speaker']['peers'])

    def test_mutiple_bgp_peers_per_bgp_speaker(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer1, \
                self.bgp_peer(name=FAKE_PEER_NAME2,
                              peer_ip=FAKE_PEER_IP2) as bgp_peer2:
            data = {"bgp_peer_id": bgp_peer1['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            data = {"bgp_peer_id": bgp_peer2['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(2, len(res['bgp_speaker']['peers']))

    def test_add_gateway_network_to_bgp_speaker_without_router(self):
        with self.bgp_speaker() as bgp_speaker:
            data = {'network_id': self._ext_net_id}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_NETWORK_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPOk.code, res.status_int)
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(self._edge_router_id,
                             res['bgp_speaker']['logical_router'])
            self.assertEqual(1, len(res['bgp_speaker']['networks']))

    def test_add_gateway_network_to_bgp_speaker_with_router(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker:
            data = {'network_id': self._ext_net_id}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_NETWORK_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_add_private_network_to_bgp_speaker_without_router(self):
        with self.bgp_speaker() as bgp_speaker, self.network() as net:
            data = {'network_id': net['network']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_NETWORK_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_add_gateway_network_without_subnets(self):
        with self.bgp_speaker() as bgp_speaker, self.network() as ext_net:
            self._set_net_external(ext_net['network']['id'])
            data = {'network_id': ext_net['network']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_NETWORK_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_add_gateway_network_with_no_gateway_ip_subnet(self):
        with self.bgp_speaker() as bgp_speaker, self.network() as ext_net:
            self._set_net_external(ext_net['network']['id'])
            self._make_subnet(self.fmt, ext_net, None, '100.64.0.0/24')
            data = {'network_id': ext_net['network']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_NETWORK_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_add_gateway_network_with_no_gateway_ip_port(self):
        with self.bgp_speaker() as bgp_speaker, self.network() as ext_net:
            self._set_net_external(ext_net['network']['id'])
            self._make_subnet(self.fmt, ext_net, '100.64.0.1', '100.64.0.0/24')
            data = {'network_id': ext_net['network']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_NETWORK_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_remove_gateway_network_from_bgp_speaker_with_peers(self):
        with self.bgp_speaker() as bgp_speaker, self.bgp_peer() as bgp_peer:
            data = {'network_id': self._ext_net_id}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_NETWORK_ACTION)
            req.get_response(self.ext_api)
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            data = {'network_id': self._ext_net_id}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          REMOVE_NETWORK_ACTION)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_remove_gateway_network_from_bgp_speaker(self):
        with self.bgp_speaker() as bgp_speaker, self.bgp_peer():
            data = {'network_id': self._ext_net_id}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_NETWORK_ACTION)
            req.get_response(self.ext_api)
            data = {'network_id': self._ext_net_id}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          REMOVE_NETWORK_ACTION)
            req.get_response(self.ext_api)
            req = self.new_show_request('bgp-speakers',
                                        bgp_speaker['bgp_speaker']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertFalse(res['bgp_speaker']['logical_router'])
            self.assertFalse(res['bgp_speaker']['networks'])

    def test_get_advertised_routes_with_single_peer(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            req = self._adver_route_list('bgp-speakers', None,
                                         bgp_speaker['bgp_speaker']['id'],
                                         'get_advertised_routes')
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(2, len(res['advertised_routes']))

    def test_get_advertised_routes_with_single_peer_and_extra_routes(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer:
            data = {"router": {"routes": [{"nexthop": self._port_fixed_ip,
                                           "destination": "20.0.0.0/8"}]}}
            req = self.new_update_request('routers',
                                          data,
                                          self._router_id)
            req.get_response(self.ext_api)
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            req = self._adver_route_list('bgp-speakers', None,
                                         bgp_speaker['bgp_speaker']['id'],
                                         'get_advertised_routes')
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(3, len(res['advertised_routes']))

    def test_get_advertised_routes_with_mutiple_peers(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer, \
                self.bgp_peer(peer_ip=FAKE_PEER_IP2) as bgp_peer2:
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            data = {"bgp_peer_id": bgp_peer2['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            req = self._adver_route_list('bgp-speakers', None,
                                         bgp_speaker['bgp_speaker']['id'],
                                         'get_advertised_routes')
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(4, len(res['advertised_routes']))

    def test_get_advertised_routes_with_mutiple_peers_and_extra_routes(self):
        with self.bgp_speaker(router_id=self._router_id) as bgp_speaker, \
                self.bgp_peer() as bgp_peer, \
                self.bgp_peer(peer_ip=FAKE_PEER_IP2) as bgp_peer2:
            data = {"router": {"routes": [{"nexthop": self._port_fixed_ip,
                                           "destination": "20.0.0.0/8"}]}}
            req = self.new_update_request('routers',
                                          data,
                                          self._router_id)
            res = req.get_response(self.ext_api)
            data = {"bgp_peer_id": bgp_peer['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            data = {"bgp_peer_id": bgp_peer2['bgp_peer']['id']}
            req = self.new_action_request('bgp-speakers', data,
                                          bgp_speaker['bgp_speaker']['id'],
                                          ADD_BGP_PEER_ACTION)
            req.get_response(self.ext_api)
            req = self._adver_route_list('bgp-speakers', None,
                                         bgp_speaker['bgp_speaker']['id'],
                                         'get_advertised_routes')
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertEqual(6, len(res['advertised_routes']))

    def _adver_route_list(self, resource, data, id, action, fmt=None,
                          subresource=None):
        return self._req(
            'GET',
            resource,
            None,
            fmt,
            id=id,
            action=action,
            subresource=subresource
        )


class BgpTestCaseML2(BgpTestCaseMixin, test_mn_ml2.MidonetPluginML2TestCase):
    pass



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# lbaasv2 driver for MidoNet

from neutron_lib import context as ncontext
from neutron_lib.plugins import directory

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_service import loopingcall

from neutron_lbaas.db.loadbalancer import models
from neutron_lbaas.drivers import driver_base
from neutron_lbaas.services.loadbalancer import constants as lb_const

from midonet.neutron.client import base as c_base
from midonet.neutron.common import config  # noqa
from midonet.neutron.common import utils

LOG = logging.getLogger(__name__)


# TODO(yamamoto): Introduce LBaaS PRECOMMIT callbacks and
# subscribe them for task-based api.


MN_STATUS_TO_OPERATING_STATUS = {
    'ACTIVE': lb_const.ONLINE,
    'INACTIVE': lb_const.OFFLINE,
    'NO_MONITOR': lb_const.NO_MONITOR,
}


class MidonetLoadBalancerDriver(driver_base.LoadBalancerBaseDriver):

    def __init__(self, plugin):
        super(MidonetLoadBalancerDriver, self).__init__(plugin)

        self.load_balancer = MidonetLoadBalancerManager(self)
        self.listener = MidonetListenerManager(self)
        self.pool = MidonetPoolManager(self)
        self.member = MidonetMemberManager(self)
        self.health_monitor = MidonetHealthMonitorManager(self)
        self._client = c_base.load_client(cfg.CONF.MIDONET)

        self.member_update_thread = loopingcall.FixedIntervalLoopingCall(
            self._update_member_status)
        self.member_update_thread.start(5, initial_delay=None,
                                        stop_on_exception=False)

        self.admin_ctx = ncontext.get_admin_context()

    def _update_member_status(self):
        # Request a list of pool members from the Midonet API
        # Go through this member list and update the neutron DB with each
        # pool member's status.
        db_members = self.plugin.db.get_pool_members(context=self.admin_ctx)
        for m in db_members:
            member_id = m.id
            midonet_member = self._client.get_pool_member(
                self.admin_ctx, member_id)
            LOG.debug("backend member: %(member)s", {'member': midonet_member})
            midonet_status = midonet_member['status']
            new_status = MN_STATUS_TO_OPERATING_STATUS[midonet_status]
            self.plugin.db.update_status(
                context=self.admin_ctx, model=models.MemberV2,
                id=member_id, operating_status=new_status)


def _build_func(method, client_method):
    if method == 'update':
        def f(self, context, old_obj, obj, **kwargs):
            return getattr(self.driver._client, client_method)(
                context=context, old_obj=old_obj, obj=obj, **kwargs)
    else:
        def f(self, context, obj, **kwargs):
            return getattr(self.driver._client, client_method)(
                context=context, obj=obj, **kwargs)
    f.__name__ = method
    return f


def _build_methods(cls, resource):
    # add methods like the following:
    #
    #    @driver_base.driver_op
    #    @log_helpers.log_method_call
    #    def <method>(self, context, obj):
    #        self._client.<method>_<resource>(context, obj)

    methods = ['create', 'update', 'delete']
    if resource == 'loadbalancerv2':
        methods += ['refresh', 'stats']
    for method in methods:
        client_method = method + '_' + resource
        f = _build_func(method, client_method)
        unbound = utils.unboundmethod(f, cls)
        wrapped = driver_base.driver_op(log_helpers.log_method_call(unbound))
        setattr(cls, method, wrapped)


def _build_manager_impl(resource):
    class _cls(object):
        pass

    _build_methods(_cls, resource)
    return _cls


class MidonetLoadBalancerManager(_build_manager_impl('loadbalancerv2'),
                                 driver_base.BaseLoadBalancerManager):

    # Override the create method here to ensure that the VIP port
    # associated with this load balancer has its 'admin_state_up' field
    # set to True. If it is set to False, midonet will not pass traffic
    # through the VIP.
    def create(self, context, obj, **kwargs):
        plugin = directory.get_plugin()
        plugin.update_port(context, obj.vip_port_id,
                           {'port': {'admin_state_up': True}})
        return super(MidonetLoadBalancerManager, self).create(
            context, obj, **kwargs)


class MidonetListenerManager(_build_manager_impl('listenerv2'),
                             driver_base.BaseListenerManager):
    pass


class MidonetPoolManager(_build_manager_impl('poolv2'),
                         driver_base.BasePoolManager):
    pass


class MidonetMemberManager(_build_manager_impl('memberv2'),
                           driver_base.BaseMemberManager):
    pass


class MidonetHealthMonitorManager(_build_manager_impl('healthmonitorv2'),
                                  driver_base.BaseHealthMonitorManager):
    pass
